import os
from PIL import ImageOps, Image

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torchvision import transforms
from torch.utils.data import Dataset, DataLoader, IterableDataset

from sklearn.model_selection import train_test_split
from sklearn.utils import shuffle

import pandas as pd

import random
import csv
import itertools
import pickle

import matplotlib.pyplot as plt
import cv2

import numpy as np
import tqdm

import pennylane as qml

from sklearn import metrics
from sklearn.metrics import classification_report

import quantum_circuits
from quantum_circuits import circuit__1

class ClassicalFrozen_QuantumTrainable__1(nn.Module):
    def __init__(self, classical_model_path, model_def):
        super().__init__()

        self.quantum_weights = {"theta_RY_switch": (3, 3),
                                "vqc_weights": (3, 1, 7, 3)}
        # self.qlayer = qml.qnn.TorchLayer(qml.transforms.broadcast_expand(circuit__1), self.quantum_weights)
        self.qlayer = qml.qnn.TorchLayer(circuit__1(bool_vals=None), self.quantum_weights)

        self.checkpoint = torch.load(classical_model_path, weights_only=True)
        self.classical_model = model_def()
        self.classical_model.load_state_dict(self.checkpoint['model'])

        for param in self.classical_model.parameters():
            param.requires_grad = False

    def forward(self, x1, x2):
        x1 = self.classical_model(x1)
        x2 = self.classical_model(x2)

        x1 = F.normalize(x1, p=2, dim=1) / np.sqrt(2)
        x2 = F.normalize(x2, p=2, dim=1) / np.sqrt(2)

        x = torch.cat((x1, x2), dim=1)
        # x = torch.stack([x1, x2], dim=0)
        # print(x1.shape, x2.shape, x.shape)
        
        overlap = self.qlayer(x)
        # overlap = torch.mean(overlap, dim=1)
        
        overlap = (1 + overlap) / 2

        # overlap = nn.Sigmoid()(overlap)

        # print(x1, x2, overlap)

        # print("overlap:", overlap)
        
        return overlap

    # def forward(self, x):
    #     x = self.classical_model(x)
        
    #     return x


import os
from PIL import ImageOps, Image

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torchvision import transforms
from torch.utils.data import Dataset, DataLoader, IterableDataset

from sklearn.model_selection import train_test_split
from sklearn.utils import shuffle

import pandas as pd

import random
import csv
import itertools
import pickle

import matplotlib.pyplot as plt
import cv2

import numpy as np
import tqdm

import pennylane as qml

from sklearn import metrics
from sklearn.metrics import classification_report

image_transform = transforms.Compose([
    transforms.Resize((155, 220)),
    ImageOps.invert,
    transforms.ToTensor(),
    # TODO: add normalize
])

img_h, img_w = 155, 220
batch_size = 128
log_interval = 50
num_epochs = 20
q_delta = 0.01
BHSig260_Hindi_path = "/kaggle/input/handwritten-signature-datasets/BHSig260-Hindi/BHSig260-Hindi/"
BHSig260_Bengali_path = "/kaggle/input/handwritten-signature-datasets/BHSig260-Bengali/BHSig260-Bengali/"
classical_branch_output_size = 256
num_channels = 2

n_qubits_single_channel = int(np.ceil(np.log2(classical_branch_output_size)))
n_qubits_switch = int(np.ceil(np.log2(num_channels)))
n_qubits = n_qubits_single_channel + n_qubits_switch
n_qubits_quantum_single_channel = lambda x : int(np.log2(x))

def visualize_sample_signature(dataset, img_w, img_h):
    '''Function to randomly select a signature from train set and
    print two genuine copies and one forged copy'''
    fig, (ax1, ax2, ax3) = plt.subplots(1, 3, figsize = (10, 10))

    # k = np.random.randint(len(BHSig260_Hindi_orig_train))
    # orig_img_names = random.sample(BHSig260_Hindi_orig_train[k], 2)
    # forg_img_name = random.sample(BHSig260_Hindi_forg_train[k], 1)

    k = np.random.randint(len(dataset))
    orig_img_names = random.sample(dataset[k], 2)
    forg_img_name = random.sample(dataset[k], 1)

    orig_img1 = cv2.imread(orig_img_names[0], 0)
    orig_img2 = cv2.imread(orig_img_names[1], 0)
    forg_img = plt.imread(forg_img_name[0], 0)
    orig_img1 = cv2.resize(orig_img1, (img_w, img_h))
    orig_img2 = cv2.resize(orig_img2, (img_w, img_h))
    forg_img = cv2.resize(forg_img, (img_w, img_h))

    ax1.imshow(orig_img1, cmap = 'gray')
    ax2.imshow(orig_img2, cmap = 'gray')
    ax3.imshow(forg_img, cmap = 'gray')

    ax1.set_title('Genuine Copy')
    ax1.axis('off')
    ax2.set_title('Genuine Copy')
    ax2.axis('off')
    ax3.set_title('Forged Copy')
    ax3.axis('off')

def generate_batch(orig_groups, forg_groups, batch_size, img_w, img_h):
    '''Function to generate a batch of data with batch_size number of data points
    Half of the data points will be Genuine-Genuine pairs and half will be Genuine-Forged pairs'''
    # img_h, img_w = 155, 220  # Assuming image height and width are 155 and 220, respectively
    while True:
        orig_pairs = []
        forg_pairs = []
        gen_gen_labels = []
        gen_for_labels = []
        all_pairs = []
        all_labels = []

        # Here we create pairs of Genuine-Genuine image names and Genuine-Forged image names
        # For every person we have 24 genuine signatures, hence we have
        # 24 choose 2 (24C2)= 276 Genuine-Genuine image pairs for one person.
        # To make Genuine-Forged pairs, we pair every Genuine signature of a person
        # with 12 randomly sampled Forged signatures of the same person.
        # Thus we make 24 * 12 = 300 Genuine-Forged image pairs for one person.
        # In all we have 120 person's data in the training data.
        # Total no. of Genuine-Genuine pairs = 120 * 276 = 33120
        # Total number of Genuine-Forged pairs = 120 * 300 = 36000
        # Total no. of data points = 33120 + 36000 = 69120
        for orig, forg in zip(orig_groups, forg_groups):
            orig_pairs.extend(list(itertools.combinations(orig, 2)))
            forg_list = list(forg)  # Ensure forg is a list
            for i in range(len(forg_list)):
                forg_pairs.extend(list(itertools.product(orig[i:i + 1], random.sample(forg_list, 12))))

        # Label for Genuine-Genuine pairs is 1
        # Label for Genuine-Forged pairs is 0
        gen_gen_labels = [1] * len(orig_pairs)
        gen_for_labels = [0] * len(forg_pairs)

        # Concatenate all the pairs together along with their labels and shuffle them
        all_pairs = orig_pairs + forg_pairs
        all_labels = gen_gen_labels + gen_for_labels
        del orig_pairs, forg_pairs, gen_gen_labels, gen_for_labels
        all_pairs, all_labels = shuffle(all_pairs, all_labels)

        k = 0
        pairs = [np.zeros((batch_size, img_h, img_w, 1), dtype=np.float32) for _ in range(2)]
        targets = np.zeros((batch_size,), dtype=np.float32)

        for ix, pair in enumerate(all_pairs):
            # Ensure pair elements are strings
            path1 = pair[0].decode('utf-8') if isinstance(pair[0], bytes) else pair[0]
            path2 = pair[1].decode('utf-8') if isinstance(pair[1], bytes) else pair[1]

            img1 = cv2.imread(path1, 0)
            img2 = cv2.imread(path2, 0)
            img1 = cv2.resize(img1, (img_w, img_h))          #Resizing Images
            img2 = cv2.resize(img2, (img_w, img_h))
            img1 = np.array(img1, dtype=np.float64)
            img2 = np.array(img2, dtype=np.float64)
            img1 /= 255                                     #Normalizing Images
            img2 /= 255
            img1 = img1[..., np.newaxis]
            img2 = img2[..., np.newaxis]
            pairs[0][k, :, :, :] = img1
            pairs[1][k, :, :, :] = img2
            targets[k] = all_labels[ix]
            k += 1
            if k == batch_size:
                yield (torch.tensor(pairs[0], dtype=torch.float32), torch.tensor(pairs[1], dtype=torch.float32)), torch.tensor(targets, dtype=torch.float32)
                k = 0
                pairs = [np.zeros((batch_size, img_h, img_w, 1), dtype=np.float32) for _ in range(2)]
                targets = np.zeros((batch_size,), dtype=np.float32)

#Creating a tf Dataset from generator
def ensure_list(groups):
    return [list(group) if not isinstance(group, list) else group for group in groups]

def make_pairs_BHSig260(orig_arr, forg_arr, filename):
    o_o = []
    o_f = []
    header = ["sign1", "sign2", "label"]
    user_idx = 0

    for user_idx in range(len(orig_arr)):
    # if True:
        for i in range(len(orig_arr[user_idx])):
            for j in range(i, len(orig_arr[user_idx])):
                o_o.append([orig_arr[user_idx][i], orig_arr[user_idx][j], 0])

    for user_idx in range(len(orig_arr)):
    # if True:
        for i in range(len(orig_arr[user_idx])):
            for j in range(len(forg_arr[user_idx])):
                o_f.append([orig_arr[user_idx][i], forg_arr[user_idx][j], 1])

    with open(filename, 'w', newline='') as f:
        writer = csv.writer(f)
        writer.writerow(header)
        writer.writerows(o_o)
        writer.writerows(o_f)

    # for user_1 in range(len(orig_arr)):
    #     for user_2 in range(user_1 + 1, len(orig_arr)):
    #         o_f = []
    #         for i in range(len(orig_arr[user_1])):
    #             for j in range(len(orig_arr[user_2])):
    #                 o_f.append([orig_arr[user_1][i], orig_arr[user_2][j], 1])

    #         with open(filename, 'a', newline='') as f:
    #             writer = csv.writer(f)
    #             writer.writerows(o_f)

def far_frr(predictions, labels, step=1e-3):
    '''Compute ROC accuracy with a range of thresholds on distances.
    '''
    dmax = np.max(predictions)
    dmin = np.min(predictions)
    nsame = np.sum(labels == 1)
    ndiff = np.sum(labels == 0)
    
    max_acc = 0
    best_thresh = -1

    thresholds=[]
    accuracies=[]
    tprr=[]
    tnrr=[]

    for d in np.arange(dmin, dmax+step, step):
        # print("d",d,"\n","dmin",dmin,"\tdmax",dmax,"\n step",step)

        idx1 = predictions.ravel() <= d
        idx2 = predictions.ravel() > d

        tpr = float(np.sum(labels[idx1] == 1)) / nsame
        tnr = float(np.sum(labels[idx2] == 0)) / ndiff
        acc = 0.5 * (tpr + tnr)
        #print ('ROC', acc, "tpr",tpr, "tnr",tnr)


        if (acc > max_acc):
            max_acc, best_thresh = acc, d

            thresholds.append(d)
            accuracies.append(acc)
            tprr.append(1-tpr)
            tnrr.append(1-tnr)

            #print("Max acc",max_acc,"\t best_thresh",best_thresh,"\n===============")

    # Create the line plot
    plt.plot(thresholds, accuracies, label = "Threshold")

    plt.xlabel('Threshold')
    plt.ylabel('Accuracy')
    plt.title('ROC Accuracy vs. Threshold')
    plt.legend()
    plt.grid(True)
    plt.show()

    plt.xlabel('Threshold')
    plt.ylabel('Accuracy')
    plt.title('FAR vs FRR')
    plt.plot(thresholds,tprr, label = "FAR")
    plt.plot(thresholds,tnrr, label = "FRR")
    plt.legend()
    plt.grid(True)
    plt.show()

    return max_acc, best_thresh,tprr,tnrr

def predict_score(model, test_dataset, device, threshold=0.5):
    '''Predict distance score and classify test images as Genuine or Forged'''

    # Get one batch of test samples
    BHSig260_Hindi_test_dataset_iter = iter(test_dataset)
    img1, img2, test_label = next(BHSig260_Hindi_test_dataset_iter)

    batch_size = img1.size(0)

    # Plot image pairs
    fig, axes = plt.subplots(batch_size, 2, figsize=(5, batch_size * 2))
    if batch_size == 1:
        axes = [axes]

    for i in range(batch_size):
        ax1, ax2 = axes[i]
        ax1.imshow(np.squeeze(img1[i].numpy()), cmap='gray')
        ax2.imshow(np.squeeze(img2[i].numpy()), cmap='gray')
        ax1.set_title("Genuine")
        ax2.set_title("Genuine" if test_label[i].item() == 1 else "Forged")
        ax1.axis("off")
        ax2.axis("off")

    plt.tight_layout()
    plt.show()

    # Inference
    img1, img2, test_label = img1.to(device), img2.to(device), test_label.to(device)
    model.eval()
    with torch.no_grad():
        out1, out2 = model(img1, img2)

        # Compute Euclidean distances
        distances = torch.norm(out1 - out2, dim=1)

        # Optionally interpret as probability with sigmoid
        probs = torch.sigmoid(distances)

    # Display scores and predicted class
    print("\nPredicted distance scores and classification:")
    for i in range(batch_size):
        score = distances[i].item()
        classification = "FORGED" if score > threshold else "GENUINE"
        print(f"Sample {i+1:02d}: Score = {score:.4f} → {classification}")


import os
from PIL import ImageOps, Image

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torchvision import transforms
from torch.utils.data import Dataset, DataLoader, IterableDataset

from sklearn.model_selection import train_test_split
from sklearn.utils import shuffle

import pandas as pd

import random
import csv
import itertools
import pickle

import matplotlib.pyplot as plt
import cv2

import numpy as np
import tqdm

import pennylane as qml

from sklearn import metrics
from sklearn.metrics import classification_report

def train_test_classical(model, optimizer, criterion, 
          train_loader, test_loader, 
          accuracy_function, distance_function,
          folder_path, num_epochs=20, log_interval=50, 
          load_checkpoint=False, start_epoch_checkpoint=0, checkpoint_start_file=None, scheduler=None):
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    device = 'mps' if torch.backends.mps.is_available() and torch.backends.mps.is_built() else device

    os.makedirs(folder_path, exist_ok=True)
    start_epoch = 1

    loss_history = []
    loss_history_eval = []
    accs_history = []
    accs_history_eval = []
    threshold_history = []

    if load_checkpoint:
        start_epoch = start_epoch_checkpoint
        checkpoint = torch.load(checkpoint_start_file, map_location=device)

        model.load_state_dict(checkpoint['model'])
        optimizer.load_state_dict(checkpoint['optim'])
        # scheduler__AmpVanilla.load_state_dict(checkpoint['scheduler'])

    for epoch in range(start_epoch, num_epochs + 1, 1):
        print(f'Epoch {epoch}/{num_epochs}')
        print('Training', '-'*20)
        print("lr:", optimizer.param_groups[0]['lr'])

        model.train()
        running_loss = 0.0
        running_accs = 0.0

        # The DataLoader now yields two images and a label per batch
        for batch_idx, (x1, x2, y) in enumerate(train_loader):
            x1, x2, y = x1.to(device), x2.to(device), y.to(device)

            optimizer.zero_grad()

            # The model's forward pass
            out1 = model(x1)
            out2 = model(x2)
            # out1, out2 = model(x1, x2)

            # The loss calculation
            loss = criterion(out1, out2, y)

            loss.backward()
            optimizer.step()

            running_loss += loss.item()

            # d_emb = torch.pairwise_distance(out1, out2, p=2)
            d_emb = distance_function(out1, out2)
            
            accs_batch, _ = accuracy_function(d_emb, y)
            running_accs += accs_batch / 100.0

            if (batch_idx + 1) % log_interval == 0:
                print(f'{batch_idx+1}/{len(train_loader)}: Loss: {running_loss / (batch_idx+1):.4f}')

        epoch_loss = running_loss / len(train_loader)
        train_accuracy = (running_accs / len(train_loader)) * 100
        print(f'Epoch {epoch}/{num_epochs}: Total training Loss: {epoch_loss:.4f} | Training Accuracy: {train_accuracy:.2f}%')

        loss_history.append(epoch_loss)
        accs_history.append(train_accuracy)

        if scheduler:
            scheduler.step()

        # After the training epoch, evaluate on the training set
        model.eval()
        running_loss = 0.0
        running_accs = 0.0
        # all_distances = []
        # all_labels = []

        dataloader = test_loader

        with torch.no_grad():
            for batch_idx, (x1, x2, y) in enumerate(dataloader):
                x1, x2, y = x1.to(device), x2.to(device), y.to(device)

                # Get the embeddings from the model
                # out1, out2 = model(x1, x2)
                out1 = model(x1)
                out2 = model(x2)

                # Calculate the loss for this batch
                loss = criterion(out1, out2, y)
                running_loss += loss.item()

                # Store distances and labels for accuracy calculation
                # distances = torch.pairwise_distance(out1, out2, p=2)
                distances = distance_function(out1, out2)

                # all_distances.append(distances.cpu())
                # all_labels.append(y.cpu())
                accs_batch, _ = accuracy_function(distances, y)
                running_accs += accs_batch / 100.0

                if (batch_idx + 1) % log_interval == 0:
                    print(f'{batch_idx+1}/{len(test_loader)}: Loss: {running_loss / (batch_idx+1):.4f}')

        # # Concatenate all distances and labels from the batches
        # all_distances = torch.cat(all_distances)
        # all_labels = torch.cat(all_labels)

        # Calculate overall loss and accuracy
        total_loss = running_loss / len(dataloader)
        # accuracy, best_threshold = accuracy_contrastive_loss(all_distances, all_labels)
        test_accuracy = (running_accs / len(dataloader)) * 100
        print(f'Epoch {epoch}/{num_epochs}: Total testing Loss: {total_loss:.4f} | Testing Accuracy: {test_accuracy:.2f}%')

        loss_history_eval.append(total_loss)
        accs_history_eval.append(test_accuracy)

        to_save = {
            'model': model.state_dict(),
            # 'scheduler': scheduler.state_dict(),
            'optim': optimizer.state_dict(),
        }

        print('Saving checkpoint and data..')
        # torch.save(to_save, '/Users/soardr/QSM/IYA Blocks/custom_Siamese_I_1to1__epoch_{}_loss_{:.3f}_acc_{:.3f}.pt'.format(epoch, loss, acc))
        torch.save(to_save, os.path.join(folder_path, f"_{epoch}.pt"))
    
    return loss_history, train_accuracy, loss_history_eval, accs_history_eval

def train_test_classical__triplet(model, optimizer, criterion, 
          train_loader, test_loader, 
          accuracy_function, distance_function,
          folder_path, num_epochs=20, log_interval=50, 
          load_checkpoint=False, start_epoch_checkpoint=0, checkpoint_start_file=None, scheduler=None):
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    device = 'mps' if torch.backends.mps.is_available() and torch.backends.mps.is_built() else device

    os.makedirs(folder_path, exist_ok=True)
    start_epoch = 1

    loss_history = []
    loss_history_eval = []
    accs_history = []
    accs_history_eval = []
    threshold_history = []

    if load_checkpoint:
        start_epoch = start_epoch_checkpoint
        checkpoint = torch.load(checkpoint_start_file, map_location=device)

        model.load_state_dict(checkpoint['model'])
        optimizer.load_state_dict(checkpoint['optim'])
        # scheduler__AmpVanilla.load_state_dict(checkpoint['scheduler'])

    for epoch in range(start_epoch, num_epochs + 1, 1):
        print(f'Epoch {epoch}/{num_epochs}')
        print('Training', '-'*20)
        print("lr:", optimizer.param_groups[0]['lr'])

        model.train()
        running_loss = 0.0
        running_accs = 0.0

        # The DataLoader now yields two images and a label per batch
        # for batch_idx, (x1, x2, y) in enumerate(train_loader):
        for batch_idx, (anchor, positive, negative) in enumerate(train_loader):
            # x1, x2, y = x1.to(device), x2.to(device), y.to(device)
            anchor, positive, negative = anchor.to(device), positive.to(device), negative.to(device)

            optimizer.zero_grad()

            # The model's forward pass
            anchor_out = model(anchor)
            positive_out = model(positive)
            negative_out = model(negative)
            # out1, out2 = model(x1, x2)

            # The loss calculation
            loss = criterion(anchor_out, positive_out, negative_out)

            loss.backward()
            optimizer.step()

            running_loss += loss.item()

            # d_emb = torch.pairwise_distance(out1, out2, p=2)
            # d_emb = distance_function(out1, out2)
            
            # accs_batch, _ = accuracy_function(d_emb, y)
            accs_batch = accuracy_function(anchor_out, positive_out, negative_out)
            
            running_accs += accs_batch / 100.0

            if (batch_idx + 1) % log_interval == 0:
                print(f'{batch_idx+1}/{len(train_loader)}: Loss: {running_loss / (batch_idx+1):.4f}')

        epoch_loss = running_loss / len(train_loader)
        train_accuracy = (running_accs / len(train_loader)) * 100
        print(f'Epoch {epoch}/{num_epochs}: Total training Loss: {epoch_loss:.4f} | Training Accuracy: {train_accuracy:.2f}%')

        loss_history.append(epoch_loss)
        accs_history.append(train_accuracy)

        if scheduler:
            scheduler.step()

        # After the training epoch, evaluate on the training set
        model.eval()
        running_loss = 0.0
        running_accs = 0.0
        # all_distances = []
        # all_labels = []

        dataloader = test_loader

        with torch.no_grad():
            # for batch_idx, (x1, x2, y) in enumerate(dataloader):
            for batch_idx, (anchor, positive, negative) in enumerate(dataloader):
                anchor, positive, negative = anchor.to(device), positive.to(device), negative.to(device)

                # Get the embeddings from the model
                # out1, out2 = model(x1, x2)
                anchor_out = model(anchor)
                positive_out = model(positive)
                negative_out = model(negative)

                # Calculate the loss for this batch
                loss = criterion(anchor_out, positive_out, negative_out)
                running_loss += loss.item()

                # Store distances and labels for accuracy calculation
                # distances = torch.pairwise_distance(out1, out2, p=2)
                # distances = distance_function(out1, out2)

                # all_distances.append(distances.cpu())
                # all_labels.append(y.cpu())
                # accs_batch, _ = accuracy_function(distances, y)
                accs_batch = accuracy_function(anchor_out, positive_out, negative_out)
                
                running_accs += accs_batch / 100.0

                if (batch_idx + 1) % log_interval == 0:
                    print(f'{batch_idx+1}/{len(test_loader)}: Loss: {running_loss / (batch_idx+1):.4f}')

        # # Concatenate all distances and labels from the batches
        # all_distances = torch.cat(all_distances)
        # all_labels = torch.cat(all_labels)

        # Calculate overall loss and accuracy
        total_loss = running_loss / len(dataloader)
        # accuracy, best_threshold = accuracy_contrastive_loss(all_distances, all_labels)
        test_accuracy = (running_accs / len(dataloader)) * 100
        print(f'Epoch {epoch}/{num_epochs}: Total testing Loss: {total_loss:.4f} | Testing Accuracy: {test_accuracy:.2f}%')

        loss_history_eval.append(total_loss)
        accs_history_eval.append(test_accuracy)

        to_save = {
            'model': model.state_dict(),
            # 'scheduler': scheduler.state_dict(),
            'optim': optimizer.state_dict(),
        }

        print('Saving checkpoint and data..')
        # torch.save(to_save, '/Users/soardr/QSM/IYA Blocks/custom_Siamese_I_1to1__epoch_{}_loss_{:.3f}_acc_{:.3f}.pt'.format(epoch, loss, acc))
        torch.save(to_save, os.path.join(folder_path, f"_{epoch}.pt"))
    
    return loss_history, train_accuracy, loss_history_eval, accs_history_eval

def train_test_quantum(model, optimizer, criterion, 
                       train_loader, test_loader, 
                       accuracy_function, distance_function,
                       folder_path, num_epochs=20, log_interval=50, 
                       load_checkpoint=False, start_epoch_checkpoint=0, checkpoint_start_file=None, scheduler=None):
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    device = 'mps' if torch.backends.mps.is_available() and torch.backends.mps.is_built() else device

    os.makedirs(folder_path, exist_ok=True)
    start_epoch = 1

    loss_history = []
    loss_history_eval = []
    accs_history = []
    accs_history_eval = []
    threshold_history = []

    if load_checkpoint:
        start_epoch = start_epoch_checkpoint
        checkpoint = torch.load(checkpoint_start_file, map_location=device)

        model.load_state_dict(checkpoint['model'])
        optimizer.load_state_dict(checkpoint['optim'])
        # scheduler__AmpVanilla.load_state_dict(checkpoint['scheduler'])

    for epoch in range(start_epoch, num_epochs + 1, 1):
        print(f'Epoch {epoch}/{num_epochs}')
        print('Training', '-'*20)
        print("lr:", optimizer.param_groups[0]['lr'])

        model.train()
        running_loss = 0.0
        running_accs = 0.0

        # The DataLoader now yields two images and a label per batch
        for batch_idx, (x1, x2, y) in enumerate(train_loader):
            x1, x2, y = x1.to(device), x2.to(device), y.to(device)

            optimizer.zero_grad()

            # The model's forward pass
            # out1 = model(x1)
            # out2 = model(x2)
            # # out1, out2 = model(x1, x2)
            overlap = model(x1, x2)

            # The loss calculation
            loss = criterion(overlap, y)

            loss.backward()
            optimizer.step()

            running_loss += loss.item()

            # d_emb = torch.pairwise_distance(out1, out2, p=2)
            if distance_function is not None:
                d_emb = distance_function(overlap)
            else:
                d_emb = overlap
            
            accs_batch, _ = accuracy_function(d_emb, y)
            # accs_batch = accuracy_function(d_emb, y)
            
            running_accs += accs_batch / 100.0

            if (batch_idx + 1) % log_interval == 0:
                print(f'{batch_idx+1}/{len(train_loader)}: Loss: {running_loss / (batch_idx+1):.4f}')

        epoch_loss = running_loss / len(train_loader)
        train_accuracy = (running_accs / len(train_loader)) * 100
        print(f'Epoch {epoch}/{num_epochs}: Total training Loss: {epoch_loss:.4f} | Training Accuracy: {train_accuracy:.2f}%')

        loss_history.append(epoch_loss)
        accs_history.append(train_accuracy)

        if scheduler:
            scheduler.step()

        # After the training epoch, evaluate on the training set
        model.eval()
        running_loss = 0.0
        running_accs = 0.0
        # all_distances = []
        # all_labels = []

        dataloader = test_loader

        with torch.no_grad():
            for batch_idx, (x1, x2, y) in enumerate(dataloader):
                x1, x2, y = x1.to(device), x2.to(device), y.to(device)

                # Get the embeddings from the model
                # # out1, out2 = model(x1, x2)
                # out1 = model(x1)
                # out2 = model(x2)
                overlap = model(x1, x2)

                # Calculate the loss for this batch
                loss = criterion(overlap, y)

                running_loss += loss.item()

                # Store distances and labels for accuracy calculation
                # distances = torch.pairwise_distance(out1, out2, p=2)
                if distance_function is not None:
                    distances = distance_function(overlap)
                else:
                    distances = overlap

                # all_distances.append(distances.cpu())
                # all_labels.append(y.cpu())
                accs_batch, _ = accuracy_function(distances, y)
                # accs_batch = accuracy_function(distances, y)
                
                running_accs += accs_batch / 100.0

                if (batch_idx + 1) % log_interval == 0:
                    print(f'{batch_idx+1}/{len(test_loader)}: Loss: {running_loss / (batch_idx+1):.4f}')

        # # Concatenate all distances and labels from the batches
        # all_distances = torch.cat(all_distances)
        # all_labels = torch.cat(all_labels)

        # Calculate overall loss and accuracy
        total_loss = running_loss / len(dataloader)
        # accuracy, best_threshold = accuracy_contrastive_loss(all_distances, all_labels)
        test_accuracy = (running_accs / len(dataloader)) * 100
        print(f'Epoch {epoch}/{num_epochs}: Total testing Loss: {total_loss:.4f} | Testing Accuracy: {test_accuracy:.2f}%')

        loss_history_eval.append(total_loss)
        accs_history_eval.append(test_accuracy)

        to_save = {
            'model': model.state_dict(),
            # 'scheduler': scheduler.state_dict(),
            'optim': optimizer.state_dict(),
        }

        print('Saving checkpoint and data..')
        # torch.save(to_save, '/Users/soardr/QSM/IYA Blocks/custom_Siamese_I_1to1__epoch_{}_loss_{:.3f}_acc_{:.3f}.pt'.format(epoch, loss, acc))
        torch.save(to_save, os.path.join(folder_path, f"_{epoch}.pt"))
    
    return loss_history, train_accuracy, loss_history_eval, accs_history_eval

def train_test_quantum__triplet(model, optimizer, criterion, 
                       train_loader, test_loader, 
                       accuracy_function, distance_function,
                       folder_path, num_epochs=20, log_interval=50, 
                       load_checkpoint=False, start_epoch_checkpoint=0, checkpoint_start_file=None, scheduler=None):
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    device = 'mps' if torch.backends.mps.is_available() and torch.backends.mps.is_built() else device

    os.makedirs(folder_path, exist_ok=True)
    start_epoch = 1

    loss_history = []
    loss_history_eval = []
    accs_history = []
    accs_history_eval = []
    threshold_history = []

    if load_checkpoint:
        start_epoch = start_epoch_checkpoint
        checkpoint = torch.load(checkpoint_start_file, map_location=device)

        model.load_state_dict(checkpoint['model'])
        optimizer.load_state_dict(checkpoint['optim'])
        # scheduler__AmpVanilla.load_state_dict(checkpoint['scheduler'])

    for epoch in range(start_epoch, num_epochs + 1, 1):
        print(f'Epoch {epoch}/{num_epochs}')
        print('Training', '-'*20)
        print("lr:", optimizer.param_groups[0]['lr'])

        model.train()
        running_loss = 0.0
        running_accs = 0.0

        # The DataLoader now yields two images and a label per batch
        # for batch_idx, (x1, x2, y) in enumerate(train_loader):
        for batch_idx, (anchor, positive, negative) in enumerate(train_loader):
            anchor, positive, negative = anchor.to(device), positive.to(device), negative.to(device)

            optimizer.zero_grad()

            # The model's forward pass
            # out1 = model(x1)
            # out2 = model(x2)
            # # out1, out2 = model(x1, x2)
            overlap_positive = model(anchor, positive)
            overlap_negative = model(anchor, negative)

            # The loss calculation
            loss = criterion(overlap_positive, overlap_negative)

            loss.backward()
            optimizer.step()

            running_loss += loss.item()

            # d_emb = torch.pairwise_distance(out1, out2, p=2)
            if distance_function is not None:
                d_emb = distance_function(overlap)
            else:
                d_emb = overlap_positive
            
            # accs_batch, _ = accuracy_function(d_emb, y)
            accs_batch = accuracy_function(anchor, overlap_positive, overlap_negative)
            
            running_accs += accs_batch / 100.0

            if (batch_idx + 1) % log_interval == 0:
                print(f'{batch_idx+1}/{len(train_loader)}: Loss: {running_loss / (batch_idx+1):.4f}')

        epoch_loss = running_loss / len(train_loader)
        train_accuracy = (running_accs / len(train_loader)) * 100
        print(f'Epoch {epoch}/{num_epochs}: Total training Loss: {epoch_loss:.4f} | Training Accuracy: {train_accuracy:.2f}%')

        loss_history.append(epoch_loss)
        accs_history.append(train_accuracy)

        if scheduler:
            scheduler.step()

        # After the training epoch, evaluate on the training set
        model.eval()
        running_loss = 0.0
        running_accs = 0.0
        # all_distances = []
        # all_labels = []

        dataloader = test_loader

        with torch.no_grad():
            # for batch_idx, (x1, x2, y) in enumerate(dataloader):
            for batch_idx, (anchor, positive, negative) in enumerate(dataloader):
                anchor, positive, negative = anchor.to(device), positive.to(device), negative.to(device)

                # Get the embeddings from the model
                # # out1, out2 = model(x1, x2)
                # out1 = model(x1)
                # out2 = model(x2)
                overlap_positive = model(anchor, positive)
                overlap_negative = model(anchor, negative)

                # Calculate the loss for this batch
                loss = criterion(overlap_positive, overlap_negative)

                running_loss += loss.item()

                # Store distances and labels for accuracy calculation
                # distances = torch.pairwise_distance(out1, out2, p=2)
                if distance_function is not None:
                    distances = distance_function(overlap)
                else:
                    distances = overlap_positive

                # all_distances.append(distances.cpu())
                # all_labels.append(y.cpu())
                # accs_batch, _ = accuracy_function(distances, y)
                accs_batch = accuracy_function(anchor, overlap_positive, overlap_negative)
                
                running_accs += accs_batch / 100.0

                if (batch_idx + 1) % log_interval == 0:
                    print(f'{batch_idx+1}/{len(test_loader)}: Loss: {running_loss / (batch_idx+1):.4f}')

        # # Concatenate all distances and labels from the batches
        # all_distances = torch.cat(all_distances)
        # all_labels = torch.cat(all_labels)

        # Calculate overall loss and accuracy
        total_loss = running_loss / len(dataloader)
        # accuracy, best_threshold = accuracy_contrastive_loss(all_distances, all_labels)
        test_accuracy = (running_accs / len(dataloader)) * 100
        print(f'Epoch {epoch}/{num_epochs}: Total testing Loss: {total_loss:.4f} | Testing Accuracy: {test_accuracy:.2f}%')

        loss_history_eval.append(total_loss)
        accs_history_eval.append(test_accuracy)

        to_save = {
            'model': model.state_dict(),
            # 'scheduler': scheduler.state_dict(),
            'optim': optimizer.state_dict(),
        }

        print('Saving checkpoint and data..')
        # torch.save(to_save, '/Users/soardr/QSM/IYA Blocks/custom_Siamese_I_1to1__epoch_{}_loss_{:.3f}_acc_{:.3f}.pt'.format(epoch, loss, acc))
        torch.save(to_save, os.path.join(folder_path, f"_{epoch}.pt"))
    
    return loss_history, train_accuracy, loss_history_eval, accs_history_eval


import os
from PIL import ImageOps, Image

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torchvision import transforms
from torch.utils.data import Dataset, DataLoader, IterableDataset

from sklearn.model_selection import train_test_split
from sklearn.utils import shuffle

import pandas as pd

import random
import csv
import itertools
import pickle

import matplotlib.pyplot as plt
import cv2

import numpy as np
import tqdm

import pennylane as qml

from sklearn import metrics
from sklearn.metrics import classification_report

import helper_funcs
from helper_funcs import q_delta, n_qubits_single_channel

import quantum_circuits
from quantum_circuits import I_Block_Vanilla__1to1, I_Block_AmplitudeEnc__variational_swap_cyclic_ansatz__w_per_pre_post_VQC, \
    AmpEmb__pre__c_post_VQC__switch_ry, circuit__1, circuit__1__AngleEmbedding

class SigNet__I_Block(nn.Module):
    def __init__(self):
        super(SigNet__I_Block, self).__init__()
        self.q_params = nn.Parameter(q_delta * torch.randn(n_qubits_single_channel))

        self.cnn1 = nn.Sequential(
            # nn.ReflectionPad2d(1),
            nn.Conv2d(1, 4, kernel_size=11),
            nn.ReLU(inplace=True),
            nn.BatchNorm2d(4),
            nn.MaxPool2d(2),

            # nn.ReflectionPad2d(1),
            nn.Conv2d(4, 4, kernel_size=5),
            nn.ReLU(inplace=True),
            nn.BatchNorm2d(4),
            nn.MaxPool2d(2),

            # nn.ReflectionPad2d(1),
            nn.Conv2d(4, 8, kernel_size=3),
            nn.ReLU(inplace=True),
            nn.BatchNorm2d(8),
            nn.MaxPool2d(2),
        )

        dummy_input = torch.zeros(1, 1, 155, 220)  # adjust size according to your input
        out = self.cnn1(dummy_input)
        flattened_size = out.view(1, -1).size(1)

        print("shape:", flattened_size)

        self.fc1 = nn.Sequential(
            nn.Linear(flattened_size, 1024),
            nn.ReLU(inplace=True),

            nn.Linear(1024, 128),
            nn.ReLU(inplace=True),

            nn.Linear(128, 4)
        )

        self.post_quantum = nn.Sequential(
            nn.Linear(4, 16),
            nn.ReLU(inplace=True),

            nn.Linear(16, 32),
            nn.ReLU(inplace=True),

            nn.Linear(32, 64)
        )

    def forward_once(self, x):
        output = self.cnn1(x)
        output = output.view(output.size()[0], -1)
        output = self.fc1(output)
        return output

    def forward(self, input1, input2):
        # Get classical feature vectors from CNN+FC pipeline
        output1 = self.forward_once(input1)  # shape: [batch_size, embedding_size]
        output2 = self.forward_once(input2)

        batch_size = output1.size(0)
        q_out_1_all = []
        q_out_2_all = []

        for i in range(batch_size):
            # Concatenate the two embeddings for this instance
            q_input_1 = output1[i]
            q_input_2 = output2[i]

            # Quantum processing (returns two tuples)
            q_out_1, q_out_2 = I_Block_Vanilla__1to1(n_qubits=8)(q_input_1, q_input_2, self.q_params, final_hadamard=True)

            # Convert from tuple to tensor and add batch dimension
            q_out_1_tensor = torch.stack(q_out_1).unsqueeze(0)  # shape: [1, n_qubits_single_channel]
            q_out_2_tensor = torch.stack(q_out_2).unsqueeze(0)

            q_out_1_all.append(q_out_1_tensor)
            q_out_2_all.append(q_out_2_tensor)

        # Combine all quantum outputs into batch tensors
        q_out_1_all = torch.cat(q_out_1_all, dim=0).float()  # shape: [batch_size, n_qubits_single_channel]
        q_out_2_all = torch.cat(q_out_2_all, dim=0).float()

        # apply post-quantum layers to calculate the contrastive loss (only one classical loss function)
        output1 = self.post_quantum(q_out_1_all)
        output2 = self.post_quantum(q_out_2_all)

        return output1, output2

    # def forward(self, input1, input2):
    #     output1 = self.forward_once(input1)
    #     output2 = self.forward_once(input2)

    #     q_out_1 = torch.Tensor(0, n_qubits_single_channel)
    #     q_out_2 = torch.Tensor(0, n_qubits_single_channel)

    #     q_out_1 = q_out_1.to(device)
    #     q_out_2 = q_out_2.to(device)

    #     for elem in q_in:
    #         q_out_elem = torch.hstack(I_Block_Vanilla__1to1(elem, elem, self.q_params)).float().unsqueeze(0)
    #         q_out = torch.cat((q_out, q_out_elem))

    #     return output1, output2

class SigNet2__I8_Block(nn.Module):
    '''
    Reference Keras: https://github.com/sounakdey/SigNet/blob/master/SigNet_v1.py
    '''
    def __init__(self):
        super().__init__()
        self.quantum_weights = {"q_weights_flat": (2*n_qubits_single_channel),
                                "pre_per_channel_VQC_weights": (2, n_qubits_single_channel, 3),
                                "post_per_channel_VQC_weights": (2, n_qubits_single_channel, 3)}
        self.qlayer = qml.qnn.TorchLayer(I_Block_AmplitudeEnc__variational_swap_cyclic_ansatz__w_per_pre_post_VQC(8), self.quantum_weights)

        self.features = nn.Sequential(
            #input size = [155, 220, 1]
            nn.Conv2d(1, 48, 11), # size = [145,210]
            nn.ReLU(),
            nn.LocalResponseNorm(size=5, k=2, alpha=1e-4, beta=0.75),
            nn.MaxPool2d(2, stride=2), # size = [72, 105]
            nn.Conv2d(48, 64, 5, padding=2, padding_mode='zeros'), # size = [72, 105]
            nn.LocalResponseNorm(size=5, k=2, alpha=1e-4, beta=0.75),
            nn.MaxPool2d(2, stride=2), # size = [36, 52]
            nn.Dropout2d(p=0.3),
            nn.Conv2d(64, 128, 3, stride=1, padding=1, padding_mode='zeros'),
            nn.Conv2d(128, 64, 3, stride=1, padding=1, padding_mode='zeros'),
            nn.MaxPool2d(2, stride=2), # size = [18, 26]
            nn.Dropout2d(p=0.3),
            nn.Flatten(1, -1), # 18*26*256
            nn.Linear(18*26*64, 1024),
            nn.Dropout2d(p=0.5),
            nn.Linear(1024, 512),
            nn.Dropout2d(p=0.5),
            nn.Linear(512, 256)
            # nn.Dropout2d(p=0.5),
            # self.qlayer,
            # nn.Linear(8, 16),
            # nn.Dropout2d(p=0.5),
            # nn.Linear(16, 32)
        )

        self.post_quantum = nn.Sequential(
            nn.Linear(8, 16),
            nn.Dropout2d(p=0.5),
            nn.Linear(16, 32)
        )

        # TODO: init bias = 0

    def forward(self, x1, x2):
        output1 = torch.tanh(self.features(x1))
        output2 = torch.tanh(self.features(x2))

        # print("output1:", output1)
        # print("output2:", output2)

        # batch_size = output1.size(0)

        # outputs_stacked = torch.stack([output1, output2], dim=0)
        # outputs_stacked = outputs_stacked.permute(1, 0, 2)
        # outputs_stacked = outputs_stacked.reshape(outputs_stacked.shape[0], outputs_stacked.shape[1]*outputs_stacked.shape[2])
        # print("shape of outputs_stacked:", outputs_stacked.shape)

        # q_out_1_all, q_out_2_all = self.qlayer(outputs_stacked)
        # # print("shape of q_layer_output:", q_out_1_all.shape, q_out_2_all.shape)

        # # Combine all quantum outputs into batch tensors
        # # q_out_1_all = torch.cat(q_out_1_all, dim=0).float()  # shape: [batch_size, n_qubits_single_channel]
        # # q_out_2_all = torch.cat(q_out_2_all, dim=0).float()

        # q_out_1_all = torch.tanh(q_out_1_all)
        # q_out_2_all = torch.tanh(q_out_2_all)

        # q_out_1_all = q_out_1_all.float()
        # q_out_2_all = q_out_2_all.float()

        # # print("q_out_1_all:", q_out_1_all)
        # # print("q_out_2_all:", q_out_2_all)

        # # apply post-quantum layers to calculate the contrastive loss (only one classical loss function)
        # output1 = self.post_quantum(q_out_1_all)
        # output2 = self.post_quantum(q_out_2_all)

        return output1, output2

class SigNet2__Amp_Vanilla__Block(nn.Module):
    '''
    Reference Keras: https://github.com/sounakdey/SigNet/blob/master/SigNet_v1.py
    '''
    def __init__(self):
        super().__init__()
        # self.quantum_weights = {"theta_RY_switch": (1),
        #                         "pre_vqc_weights": (2, n_qubits_single_channel, 3),
        #                         "post_vqc_weights": (2, n_qubits_single_channel, 3)}
        self.quantum_weights = {"pre_vqc_weights_A": (2, n_qubits_single_channel, 3),
                                "pre_vqc_weights_B": (2, n_qubits_single_channel, 3),
                                "post_vqc_weights_A": (2, n_qubits_single_channel, 3),
                                "post_vqc_weights_B": (2, n_qubits_single_channel, 3),
                                "pre_vqc_weights_C": (2, n_qubits_single_channel, 3),
                                "pre_vqc_weights_D": (2, n_qubits_single_channel, 3),
                                "post_vqc_weights_C": (2, n_qubits_single_channel, 3),
                                "post_vqc_weights_D": (2, n_qubits_single_channel, 3)}

        # self.qlayer = qml.qnn.TorchLayer(AmpEmb__Vanilla__pre_post_VQC__switch_ry, self.quantum_weights)
        self.qlayer = qml.qnn.TorchLayer(AmpEmb__pre__c_post_VQC__switch_ry(8), self.quantum_weights)

        self.features = nn.Sequential(
            #input size = [155, 220, 1]
            nn.Conv2d(1, 32, 11), # size = [145,210]
            nn.ReLU(),
            nn.LocalResponseNorm(size=5, k=2, alpha=1e-4, beta=0.75),
            nn.MaxPool2d(2, stride=2), # size = [72, 105]
            nn.Conv2d(32, 48, 5, padding=2, padding_mode='zeros'), # size = [72, 105]
            nn.LocalResponseNorm(size=5, k=2, alpha=1e-4, beta=0.75),
            nn.MaxPool2d(2, stride=2), # size = [36, 52]
            nn.Dropout2d(p=0.3),
            nn.Conv2d(48, 64, 3, stride=1, padding=1, padding_mode='zeros'),
            nn.Conv2d(64, 48, 3, stride=1, padding=1, padding_mode='zeros'),
            nn.MaxPool2d(2, stride=2), # size = [18, 26]
            nn.Dropout2d(p=0.3),
            nn.Flatten(1, -1), # 18*26*256
            nn.Linear(18*26*48, 2048),
            nn.Dropout2d(p=0.5),
            nn.Linear(2048, 1024)
            # nn.Dropout2d(p=0.5),
            # nn.Linear(1024, 512)
            # nn.Dropout2d(p=0.5),
            # self.qlayer,
            # nn.Linear(8, 16),
            # nn.Dropout2d(p=0.5),
            # nn.Linear(16, 32)
        )

        self.post_quantum = nn.Sequential(
            nn.Linear(8, 16),
            nn.Dropout2d(p=0.5),
            nn.Linear(16, 32)
        )

        # TODO: init bias = 0

    def forward(self, x1, x2):
        # output1 = self.features(x1)
        # output2 = self.features(x2)

        output1 = torch.tanh(self.features(x1))
        output2 = torch.tanh(self.features(x2))

        # print("output1:", output1)
        # print("output2:", output2)

        batch_size = output1.size(0)

        output1 = F.normalize(output1, p=2, dim=1) / np.sqrt(2)
        output2 = F.normalize(output2, p=2, dim=1) / np.sqrt(2)

        outputs_stacked = torch.stack([output1, output2], dim=0)
        outputs_stacked = outputs_stacked.permute(1, 0, 2)
        outputs_stacked = outputs_stacked.reshape(outputs_stacked.shape[0], outputs_stacked.shape[1]*outputs_stacked.shape[2])
        # print("shape of outputs_stacked:", outputs_stacked.shape)

        state_overlap = self.qlayer(outputs_stacked)
        # print("shape of q_layer_output:", q_out_1_all.shape, q_out_2_all.shape)

        # # Combine all quantum outputs into batch tensors
        # # q_out_1_all = torch.cat(q_out_1_all, dim=0).float()  # shape: [batch_size, n_qubits_single_channel]
        # # q_out_2_all = torch.cat(q_out_2_all, dim=0).float()

        # q_out_1_all = torch.tanh(q_out_1_all)
        # q_out_2_all = torch.tanh(q_out_2_all)

        # q_out_1_all = q_out_1_all.float()
        # q_out_2_all = q_out_2_all.float()

        # # print("q_out_1_all:", q_out_1_all)
        # # print("q_out_2_all:", q_out_2_all)

        # # apply post-quantum layers to calculate the contrastive loss (only one classical loss function)
        # output1 = self.post_quantum(q_out_1_all)
        # output2 = self.post_quantum(q_out_2_all)

        # return output1, output2

        return state_overlap

class SigNet__Circuit_1(nn.Module):
    def __init__(self):
        super().__init__()
        self.quantum_weights = {"theta_RY_switch": (3, 3),
                                "vqc_weights": (3, 2, 7, 3)}

        self.qlayer = qml.qnn.TorchLayer(circuit__1(bool_vals=None), self.quantum_weights)

        self.features = nn.Sequential(
            # Input: [1, 155, 220]
            nn.Conv2d(1, 96, kernel_size=11, stride=4),  # [96, 37, 53]
            nn.ReLU(),
            nn.LocalResponseNorm(size=5, k=2, alpha=1e-4, beta=0.75),
            nn.MaxPool2d(kernel_size=3, stride=2),  # [96, 18, 26]
            nn.Conv2d(96, 256, kernel_size=5, padding=2),  # [256, 18, 26]
            nn.LocalResponseNorm(size=5, k=2, alpha=1e-4, beta=0.75),
            nn.MaxPool2d(kernel_size=3, stride=2),  # [256, 8, 12]
            nn.Dropout2d(p=0.3),
            nn.Conv2d(256, 384, kernel_size=3, padding=1),  # [384, 8, 12]
            nn.Conv2d(384, 256, kernel_size=3, padding=1),  # [256, 8, 12]
            nn.MaxPool2d(kernel_size=3, stride=2),  # [256, 3, 5]
            nn.Dropout2d(p=0.3),
            nn.Flatten(),  # 256*3*5 = 3,840
            nn.Linear(3840, 1024),
            nn.Dropout(p=0.5),  # Note: Dropout2d → Dropout for Linear layers
            nn.Linear(1024, 256),
        )

    def forward(self, x1, x2):
        x1 = self.features(x1)
        x2 = self.features(x2)

        x1 = F.normalize(x1, p=2, dim=1) / np.sqrt(2)
        x2 = F.normalize(x2, p=2, dim=1) / np.sqrt(2)

        quantum_input = torch.cat((x1, x2), dim=1)
        quantum_output = self.qlayer(quantum_input)

        # final_out = nn.Sigmoid()(quantum_output)

        return (1 + quantum_output) / 2
        # return final_out

        # return x1, x2

class SigNet_Half__Circuit_1(nn.Module):
    def __init__(self):
        super().__init__()
        self.quantum_weights = {"theta_RY_switch": (3, 3),
                                "vqc_weights": (3, 1, 7, 3)}

        self.qlayer = qml.qnn.TorchLayer(qml.transforms.broadcast_expand(circuit__1(bool_vals=None)), self.quantum_weights)

        self.features = nn.Sequential(
            # Input: [1, 155, 220]
            nn.Conv2d(1, 48, kernel_size=11, stride=4),  # [96, 37, 53]
            nn.ReLU(),
            nn.LocalResponseNorm(size=5, k=2, alpha=1e-4, beta=0.75),
            nn.MaxPool2d(kernel_size=3, stride=2),  # [96, 18, 26]
            nn.Conv2d(48, 128, kernel_size=5, padding=2),  # [256, 18, 26]
            nn.LocalResponseNorm(size=5, k=2, alpha=1e-4, beta=0.75),
            nn.MaxPool2d(kernel_size=3, stride=2),  # [256, 8, 12]
            nn.Dropout2d(p=0.3),
            nn.Conv2d(128, 192, kernel_size=3, padding=1),  # [384, 8, 12]
            nn.Conv2d(192, 128, kernel_size=3, padding=1),  # [256, 8, 12]
            nn.MaxPool2d(kernel_size=3, stride=2),  # [256, 3, 5]
            nn.Dropout2d(p=0.3),
            nn.Flatten(),  # 256*3*5 = 3,840
            nn.Linear(1920, 512),
            nn.Dropout(p=0.5),  # Note: Dropout2d → Dropout for Linear layers
            nn.Linear(512, 128),
        )

    def forward(self, x1, x2):
        x1 = self.features(x1)
        x2 = self.features(x2)

        x1 = F.normalize(x1, p=2, dim=1) / np.sqrt(2)
        x2 = F.normalize(x2, p=2, dim=1) / np.sqrt(2)

        quantum_input = torch.cat((x1, x2), dim=1)
        quantum_output = self.qlayer(quantum_input)

        # final_out = nn.Sigmoid()(quantum_output)
        # print("quantum_output:", quantum_output)

        return (1 + quantum_output) / 2
        # return final_out

        # return x1, x2

class SigNet_Half__Circuit_1__AngEnc(nn.Module):
    def __init__(self):
        super().__init__()
        self.quantum_weights = {"theta_RY_switch": (3, 3),
                                "vqc_weights": (3, 2, 8, 3)}

        self.qlayer = qml.qnn.TorchLayer(circuit__1__AngleEmbedding(), self.quantum_weights)

        self.features = nn.Sequential(
            # Input: [1, 155, 220]
            nn.Conv2d(1, 48, kernel_size=11, stride=4),  # [96, 37, 53]
            nn.ReLU(),
            nn.LocalResponseNorm(size=5, k=2, alpha=1e-4, beta=0.75),
            nn.MaxPool2d(kernel_size=3, stride=2),  # [96, 18, 26]
            nn.Conv2d(48, 128, kernel_size=5, padding=2),  # [256, 18, 26]
            nn.LocalResponseNorm(size=5, k=2, alpha=1e-4, beta=0.75),
            nn.MaxPool2d(kernel_size=3, stride=2),  # [256, 8, 12]
            nn.Dropout2d(p=0.3),
            nn.Conv2d(128, 192, kernel_size=3, padding=1),  # [384, 8, 12]
            nn.Conv2d(192, 128, kernel_size=3, padding=1),  # [256, 8, 12]
            nn.MaxPool2d(kernel_size=3, stride=2),  # [256, 3, 5]
            nn.Dropout2d(p=0.3),
            nn.Flatten(),  # 256*3*5 = 3,840
            nn.Linear(1920, 512),
            nn.Dropout(p=0.5),  # Note: Dropout2d → Dropout for Linear layers
            nn.Linear(512, 7),
        )

    def forward(self, x1, x2):
        x1 = self.features(x1)
        x2 = self.features(x2)

        # x1 = F.normalize(x1, p=2, dim=1) / np.sqrt(2)
        # x2 = F.normalize(x2, p=2, dim=1) / np.sqrt(2)
        x1 = F.normalize(x1, p=2, dim=1) * np.pi
        x2 = F.normalize(x2, p=2, dim=1) * np.pi

        quantum_input = torch.cat((x1, x2), dim=1)
        quantum_output = self.qlayer(quantum_input)

        # final_out = nn.Sigmoid()(quantum_output)

        return (1 + quantum_output) / 2
        # return final_out

        # return x1, x2


import os
from PIL import ImageOps, Image

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torchvision import transforms
from torch.utils.data import Dataset, DataLoader, IterableDataset

from sklearn.model_selection import train_test_split
from sklearn.utils import shuffle

import pandas as pd

import random
import csv
import itertools
import pickle

import matplotlib.pyplot as plt
import cv2

import numpy as np
import tqdm

import pennylane as qml

from sklearn import metrics
from sklearn.metrics import classification_report

class ContrastiveLoss(nn.Module):
    def __init__(self, alpha, beta, margin):
        super().__init__()
        self.alpha = alpha
        self.beta = beta
        self.margin = margin

    def forward(self, x1, x2, y):
        '''
        Shapes:
        -------
        x1: [B,C]
        x2: [B,C]
        y: [B,1]

        Returns:
        --------
        loss: [B,1]]
        '''
        distance = torch.pairwise_distance(x1, x2, p=2)
        # print("shapes:", x1.shape, x2.shape, y.shape)
        loss = self.alpha * (1-y) * distance**2 + \
               self.beta * y * (torch.max(torch.zeros_like(distance), self.margin - distance)**2)
        return torch.mean(loss, dtype=torch.float)

class TripletMarginLoss(nn.Module):
    """
    A custom loss function for triplet networks. It aims to pull the anchor
    and positive embeddings closer together while pushing the anchor and negative
    embeddings apart by at least a specified margin.
    """
    def __init__(self, margin=1.0, p=2):
        """
        Args:
            margin (float): The minimum distance that the anchor-negative pair
                            must be from the anchor-positive pair.
            p (int): The norm degree for the distance calculation (e.g., p=2 for Euclidean distance).
        """
        super(TripletMarginLoss, self).__init__()
        self.margin = margin
        self.p = p

    def forward(self, anchor, positive, negative):
        """
        Calculates the triplet loss.

        Args:
            anchor (torch.Tensor): The embedding vector for the anchor images.
            positive (torch.Tensor): The embedding vector for the positive images.
            negative (torch.Tensor): The embedding vector for the negative images.

        Returns:
            torch.Tensor: The mean loss for the batch.
        """
        # Calculate the distance between the anchor and positive embeddings
        distance_positive = F.pairwise_distance(anchor, positive, p=self.p)

        # Calculate the distance between the anchor and negative embeddings
        distance_negative = F.pairwise_distance(anchor, negative, p=self.p)

        # Calculate the loss
        # loss = max(0, distance_positive - distance_negative + margin)
        loss = torch.relu(distance_positive - distance_negative + self.margin)

        return torch.mean(loss)

class TripletMarginLoss__Quantum(nn.Module):
    """
    A custom loss function for triplet networks. It aims to pull the anchor
    and positive embeddings closer together while pushing the anchor and negative
    embeddings apart by at least a specified margin.
    """
    def __init__(self, margin=1.0, p=2):
        """
        Args:
            margin (float): The minimum distance that the anchor-negative pair
                            must be from the anchor-positive pair.
            p (int): The norm degree for the distance calculation (e.g., p=2 for Euclidean distance).
        """
        super(TripletMarginLoss__Quantum, self).__init__()
        self.margin = margin
        self.p = p

    # def forward(self, anchor, positive, negative):
    def forward(self, positive_dist, negative_dist):
        """
        Calculates the triplet loss.

        Args:
            anchor (torch.Tensor): The embedding vector for the anchor images.
            positive (torch.Tensor): The embedding vector for the positive images.
            negative (torch.Tensor): The embedding vector for the negative images.

        Returns:
            torch.Tensor: The mean loss for the batch.
        """
        # # Calculate the distance between the anchor and positive embeddings
        # distance_positive = F.pairwise_distance(anchor, positive, p=self.p)

        # # Calculate the distance between the anchor and negative embeddings
        # distance_negative = F.pairwise_distance(anchor, negative, p=self.p)

        # Calculate the loss
        # loss = max(0, distance_positive - distance_negative + margin)
        loss = torch.relu(0 - positive_dist + negative_dist + self.margin)

        return torch.mean(loss)

class ContrastiveLoss_Quantum(nn.Module):
    def __init__(self, alpha, beta, margin):
        super().__init__()
        self.alpha = alpha
        self.beta = beta
        self.margin = margin

    def forward(self, overlap, y):
        '''
        Shapes:
        -------
        x1: [B,C]
        x2: [B,C]
        y: [B,1]

        Returns:
        --------
        loss: [B,1]]
        '''
        distance = (1 - overlap) / 2

        loss = self.alpha * (1-y) * distance**2 + \
               self.beta * y * (torch.max(torch.zeros_like(distance), self.margin - distance)**2)
        return torch.mean(loss, dtype=torch.float)

class ContrastiveLoss_Quantum_Sigmoid(nn.Module):
    def __init__(self, alpha, beta, margin):
        super().__init__()
        self.alpha = alpha
        self.beta = beta
        self.margin = margin

    def forward(self, overlap, y):
        '''
        Shapes:
        -------
        x1: [B,C]
        x2: [B,C]
        y: [B,1]

        Returns:
        --------
        loss: [B,1]]
        '''
        # distance = torch.pairwise_distance(x1, x2, p=2)
        distance = overlap
        
        loss = self.alpha * (1-y) * distance**2 + \
               self.beta * y * (torch.max(torch.zeros_like(distance), self.margin - distance)**2)
        return torch.mean(loss, dtype=torch.float)

class ContrastiveLoss_Quantum_Direct(nn.Module):
    def __init__(self, alpha, beta, margin, overlap_func=None):
        super().__init__()
        self.alpha = alpha
        self.beta = beta
        self.margin = margin
        self.overlap_func = overlap_func

    def forward(self, overlap, y):
        '''
        Shapes:
        -------
        x1: [B,C]
        x2: [B,C]
        y: [B,1]

        Returns:
        --------
        loss: [B,1]]
        '''
        # distance = torch.pairwise_distance(x1, x2, p=2)
        if self.overlap_func is None:
            distance = overlap
            # 0: different class
            # 1: same class
        else:
            distance = self.overlap_func(overlap)
        
        # loss = self.alpha * (1-y) * (1 - distance)**2 + \
        #        self.beta * y * (torch.max(torch.zeros_like(distance), self.margin - 1 + distance)**2)

        loss_similar = self.alpha * (1 - overlap)**2
        loss_dissimilar = self.beta * torch.relu(overlap - self.margin)**2
        loss = (1 - y) * loss_similar + y * loss_dissimilar
        
        return torch.mean(loss, dtype=torch.float)

class ContrastiveLoss_Quantum_Direct_2(nn.Module):
    """
    Implements a custom contrastive loss function adapted for a similarity
    score (overlap) between 0 and 1.

    The loss is defined as: L = (1-y)*(1-overlap)^2 + y*(overlap)^2
    where y=0 for similar pairs and y=1 for dissimilar pairs.
    """
    def __init__(self):
        super().__init__()

    def forward(self, overlap, y):
        """
        Calculates the contrastive loss for a batch of pairs using the overlap score.
        
        Args:
            overlap (torch.Tensor): A 1D tensor of overlap scores (0-1).
            y (torch.Tensor): A 1D tensor of ground-truth labels (0 for similar, 1 for dissimilar).
            
        Returns:
            torch.Tensor: The mean loss for the batch.
        """
        # Ensure y has the same dimensions as overlap for element-wise operations
        if y.ndim != overlap.ndim:
            y = y.view_as(overlap)

        # Loss for similar pairs (y=0) - minimizes (1 - overlap)^2,
        # encouraging overlap to be close to 1.
        loss_similar = (1 - y) * torch.pow(1 - overlap, 2)
        
        # Loss for dissimilar pairs (y=1) - minimizes overlap^2,
        # encouraging overlap to be close to 0.
        loss_dissimilar = y * torch.pow(overlap, 2)
        
        # The total loss is the average of the combined components
        loss = torch.mean(loss_similar + loss_dissimilar)
        
        return loss

class BCEOverlapLoss_Quantum(nn.Module):
    def __init__(self):
        super().__init__()
        self.loss_fn = nn.BCEWithLogitsLoss()

    def forward(self, overlap, y):
        # Map overlap from [-1, 1] → logits in R
        logits = overlap * 3  # scale to increase contrast
        return self.loss_fn(logits.squeeze(), y.float())


import os
from PIL import ImageOps, Image

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torchvision import transforms
from torch.utils.data import Dataset, DataLoader, IterableDataset

from sklearn.model_selection import train_test_split
from sklearn.utils import shuffle

import pandas as pd

import random
import csv
import itertools
import pickle

import matplotlib.pyplot as plt
import cv2

import numpy as np
import tqdm

import pennylane as qml

from sklearn import metrics
from sklearn.metrics import classification_report

class SignDataset(Dataset):
    def __init__(self, is_train: bool, data_dir: str, image_transform=None):
        if not os.path.exists(os.path.join(data_dir, 'train.csv')) or not os.path.exists(os.path.join(data_dir, 'test.csv')):
            print('Not found train/test splits, run create_annotation first')
        else:
            print('Use existed train/test splits')

        if is_train:
            self.df = pd.read_csv(os.path.join(data_dir, 'train.csv'), header=None)
        else:
            self.df = pd.read_csv(os.path.join(data_dir, 'test.csv'), header=None)

        self.image_transform = image_transform

    def __len__(self):
        return len(self.df)

    def __getitem__(self, index):
        x1, x2, y = self.df.iloc[index]

        x1 = Image.open(x1).convert('L')
        x2 = Image.open(x2).convert('L')

        if self.image_transform:
            x1 = self.image_transform(x1)
            x2 = self.image_transform(x2)

        return x1, x2, y

class CSVDataset(Dataset):
    """
    A custom PyTorch Dataset for loading image pairs and labels from a CSV file.
    """
    def __init__(self, csv_file, transform=None):
        """
        Args:
            csv_file (string): Path to the csv file with pairs and labels.
            transform (callable, optional): Optional transform to be applied on a sample.
        """
        # Read the CSV file using pandas
        self.pairs_df = pd.read_csv(csv_file)
        self.transform = transform

    def __len__(self):
        # The number of samples is the number of rows in the CSV file
        return len(self.pairs_df)

    def __getitem__(self, idx):
        # Get the file paths and label from the dataframe row
        row = self.pairs_df.iloc[idx]
        img1_path = row['sign1']
        img2_path = row['sign2']
        label = row['label']

        # Load the images
        # 'L' converts to grayscale, which is common for signature analysis
        image1 = Image.open(img1_path).convert('L')
        image2 = Image.open(img2_path).convert('L')

        # Apply transformations if provided
        if self.transform:
            image1 = self.transform(image1)
            image2 = self.transform(image2)

        # Convert label to a PyTorch tensor
        label = torch.tensor(label, dtype=torch.float32)

        return image1, image2, label

class TripletDataset(Dataset):
    def __init__(self, org_signs, forg_signs, transform=None):
        self.org_data = org_signs
        self.forg_data = forg_signs
        self.signer_ids = [i for i in range(len(org_signs))]
        self.transform = transform

    def __len__(self):
        # We'll use a large number for the length to ensure we get enough batches.
        return len(self.signer_ids) * 1000

    def __getitem__(self, index):
        # 1. Select a random anchor signer
        anchor_signer_id = random.choice(self.signer_ids)

        # anchor_signatures = self.signer_data[anchor_signer_id]['orig']
        anchor_signatures = self.org_data[anchor_signer_id]

        # 2. Select the anchor and positive images from the same signer
        anchor_path, positive_path = random.sample(anchor_signatures, 2)

        # 3. Select a negative image
        # To make the model robust, we can choose a negative from either a different
        # signer's genuine signature or a forgery of the anchor signer.

        if random.random() < 0.5 and len(self.forg_data[anchor_signer_id]) > 0:
            # Case A: Negative is a forgery of the anchor signer
            negative_path = random.choice(self.forg_data[anchor_signer_id])
        else:
            # Case B: Negative is a genuine signature from a different signer
            negative_signer_id = random.choice([sid for sid in self.signer_ids if sid != anchor_signer_id])
            negative_signatures = self.org_data[negative_signer_id]
            negative_path = random.choice(negative_signatures)

        # Load images and apply transformations
        anchor_image = Image.open(anchor_path).convert('L')
        positive_image = Image.open(positive_path).convert('L')
        negative_image = Image.open(negative_path).convert('L')

        if self.transform:
            anchor_image = self.transform(anchor_image)
            positive_image = self.transform(positive_image)
            negative_image = self.transform(negative_image)

        # Return a triplet of images
        return anchor_image, positive_image, negative_image

class TripletDataset_CL(Dataset):
    """
    A custom PyTorch Dataset that dynamically generates pairs of signatures
    (positive and negative) for training a Siamese network with contrastive loss.
    """
    def __init__(self, org_signs, forg_signs, transform=None):
        """
        Args:
            org_signs (list): A list of lists, where each sublist contains the
                              file paths of original signatures for one signer.
            forg_signs (list): A list of lists, where each sublist contains the
                               file paths of forged signatures for one signer.
            transform (callable, optional): Optional transform to be applied on a sample.
        """
        self.org_data = org_signs
        self.forg_data = forg_signs
        self.signer_ids = [i for i in range(len(org_signs))]
        self.transform = transform

    def __len__(self):
        # We'll use a large number for the length to ensure we get enough batches.
        # This number should be large enough to run for a full epoch.
        return len(self.signer_ids) * 1000

    def __getitem__(self, index):
        # Dynamically generate a positive or negative pair with a 50% chance
        is_positive = random.random() < 0.5

        if is_positive:
            # Create a positive pair (label 0)
            signer_id = random.choice(self.signer_ids)
            signatures = self.org_data[signer_id]

            # Select two different original signatures from the same signer
            # Ensure there are at least two signatures to sample from
            if len(signatures) < 2:
                # Handle cases with too few signatures by falling back to a different signer
                return self.__getitem__(random.randint(0, self.__len__() - 1))

            img1_path, img2_path = random.sample(signatures, 2)
            label = 0

        else:
            # Create a negative pair (label 1)
            # To make the model robust, we mix between genuine-forged and genuine-genuine pairs
            if random.random() < 0.5:
                # Case A: Genuine-forged pair from a single signer
                signer_id = random.choice(self.signer_ids)
                orig_sigs = self.org_data[signer_id]
                forg_sigs = self.forg_data[signer_id]

                # Ensure both lists are not empty
                if not orig_sigs or not forg_sigs:
                    return self.__getitem__(random.randint(0, self.__len__() - 1))

                img1_path = random.choice(orig_sigs)
                img2_path = random.choice(forg_sigs)
            else:
                # Case B: Genuine-genuine pair from two different signers
                signer_id1, signer_id2 = random.sample(self.signer_ids, 2)
                orig_sigs1 = self.org_data[signer_id1]
                orig_sigs2 = self.org_data[signer_id2]

                if not orig_sigs1 or not orig_sigs2:
                    return self.__getitem__(random.randint(0, self.__len__() - 1))

                img1_path = random.choice(orig_sigs1)
                img2_path = random.choice(orig_sigs2)

            label = 1

        # Load images and apply transformations
        img1 = Image.open(img1_path).convert('L')
        img2 = Image.open(img2_path).convert('L')

        if self.transform:
            img1 = self.transform(img1)
            img2 = self.transform(img2)

        # The label for contrastive loss should be a float tensor
        label_tensor = torch.tensor(label, dtype=torch.float32)

        return img1, img2, label_tensor

class SiameseBHSigDataset(IterableDataset):
    def __init__(self, orig_groups, forg_groups, batch_size=32, img_h=155, img_w=220):
        self.orig_groups = [list(g) for g in orig_groups]
        self.forg_groups = [list(g) for g in forg_groups]
        self.batch_size = batch_size
        self.img_h = img_h
        self.img_w = img_w

    def generate_batches(self):
        while True:
        # if True:
            orig_pairs = []
            forg_pairs = []

            for orig, forg in zip(self.orig_groups, self.forg_groups):
                # Genuine-Genuine pairs (24C2 = 276 per person)
                orig_pairs.extend(list(itertools.combinations(orig, 2)))
                # Genuine-Forged pairs (24 * 12 = 300 per person)
                for i in range(len(orig)):
                    forg_pairs.extend([(orig[i], f) for f in random.sample(forg, 12)])

            all_pairs = orig_pairs + forg_pairs
            all_labels = [1] * len(orig_pairs) + [0] * len(forg_pairs)
            all_pairs, all_labels = shuffle(all_pairs, all_labels)

            k = 0
            pairs = [np.zeros((self.batch_size, self.img_h, self.img_w, 1), dtype=np.float32) for _ in range(2)]
            targets = np.zeros((self.batch_size,), dtype=np.float32)

            print("length of all_pairs:", len(all_pairs))
            print("length of all_labels:", len(all_labels))

            for (img1_path, img2_path), label in zip(all_pairs, all_labels):
                # Load and preprocess images
                img1 = cv2.imread(img1_path, 0)
                img2 = cv2.imread(img2_path, 0)
                img1 = cv2.resize(img1, (self.img_w, self.img_h)) / 255.0
                img2 = cv2.resize(img2, (self.img_w, self.img_h)) / 255.0
                img1 = img1[..., np.newaxis]
                img2 = img2[..., np.newaxis]

                pairs[0][k] = img1
                pairs[1][k] = img2
                targets[k] = label
                k += 1

                if k == self.batch_size:
                    # Convert to PyTorch tensors
                    x1 = torch.tensor(pairs[0], dtype=torch.float32).permute(0, 3, 1, 2)  # (B, C, H, W)
                    x2 = torch.tensor(pairs[1], dtype=torch.float32).permute(0, 3, 1, 2)
                    y = torch.tensor(targets, dtype=torch.float32)
                    yield x1, x2, y
                    # Reset for next batch
                    k = 0
                    pairs = [np.zeros((self.batch_size, self.img_h, self.img_w, 1), dtype=np.float32) for _ in range(2)]
                    targets = np.zeros((self.batch_size,), dtype=np.float32)

    def __iter__(self):
        return iter(self.generate_batches())


from setuptools import setup

setup(
    name='qsm_soham',
    version='0.1.0',    
    description='QSM',
    url='https://github.com/SoardRaspi/QSM',
    author='Soham Pawar',
    author_email='Soham.Pawar127@iiitb.ac.in',
    license='GNU v3',
    packages=['qsm'],
    install_requires=['torch', 'torchmedia', 'torchvision', 'matplotlib', 'pandas', 'numpy', 'opencv-python', 'scikit-learn', 'pennylane',],

    # classifiers=[
    #     'Development Status :: 1 - Planning',
    #     'Intended Audience :: Science/Research',
    #     'License :: OSI Approved :: BSD License',  
    #     'Operating System :: POSIX :: Linux',        
    #     'Programming Language :: Python :: 2',
    #     'Programming Language :: Python :: 2.7',
    #     'Programming Language :: Python :: 3',
    #     'Programming Language :: Python :: 3.4',
    #     'Programming Language :: Python :: 3.5',
    # ],
)


import os
from PIL import ImageOps, Image

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torchvision import transforms
from torch.utils.data import Dataset, DataLoader, IterableDataset

from sklearn.model_selection import train_test_split
from sklearn.utils import shuffle

import pandas as pd

import random
import csv
import itertools
import pickle

import matplotlib.pyplot as plt
import cv2

import numpy as np
import tqdm

import pennylane as qml

from sklearn import metrics
from sklearn.metrics import classification_report

class SigNet__Half(nn.Module):
    def __init__(self):
        super().__init__()
        self.features = nn.Sequential(
            # Input: [1, 155, 220]
            nn.Conv2d(1, 48, kernel_size=11, stride=4),  # [96, 37, 53]
            nn.ReLU(),
            nn.LocalResponseNorm(size=5, k=2, alpha=1e-4, beta=0.75),
            nn.MaxPool2d(kernel_size=3, stride=2),  # [96, 18, 26]
            nn.Conv2d(48, 128, kernel_size=5, padding=2),  # [256, 18, 26]
            nn.LocalResponseNorm(size=5, k=2, alpha=1e-4, beta=0.75),
            nn.MaxPool2d(kernel_size=3, stride=2),  # [256, 8, 12]
            nn.Dropout2d(p=0.3),
            nn.Conv2d(128, 192, kernel_size=3, padding=1),  # [384, 8, 12]
            nn.Conv2d(192, 128, kernel_size=3, padding=1),  # [256, 8, 12]
            nn.MaxPool2d(kernel_size=3, stride=2),  # [256, 3, 5]
            nn.Dropout2d(p=0.3),
            nn.Flatten(),  # 256*3*5 = 3,840
            nn.Linear(1920, 512),
            nn.Dropout(p=0.5),  # Note: Dropout2d → Dropout for Linear layers
            nn.Linear(512, 128),
        )

    def forward(self, x1, x2):
        x1 = self.features(x1)
        x2 = self.features(x2)
        return x1, x2

class SigNet__Half_Single(nn.Module):
    def __init__(self, classical_final_output=128):
        super().__init__()
        self.features = nn.Sequential(
            # Input: [1, 155, 220]
            nn.Conv2d(1, 48, kernel_size=11, stride=4),  # [96, 37, 53]
            nn.ReLU(),
            nn.LocalResponseNorm(size=5, k=2, alpha=1e-4, beta=0.75),
            nn.MaxPool2d(kernel_size=3, stride=2),  # [96, 18, 26]
            nn.Conv2d(48, 128, kernel_size=5, padding=2),  # [256, 18, 26]
            nn.LocalResponseNorm(size=5, k=2, alpha=1e-4, beta=0.75),
            nn.MaxPool2d(kernel_size=3, stride=2),  # [256, 8, 12]
            nn.Dropout2d(p=0.3),
            nn.Conv2d(128, 192, kernel_size=3, padding=1),  # [384, 8, 12]
            nn.Conv2d(192, 128, kernel_size=3, padding=1),  # [256, 8, 12]
            nn.MaxPool2d(kernel_size=3, stride=2),  # [256, 3, 5]
            nn.Dropout2d(p=0.3),
            nn.Flatten(),  # 256*3*5 = 3,840
            nn.Linear(1920, 512),
            nn.Dropout(p=0.5),  # Note: Dropout2d → Dropout for Linear layers
            nn.Linear(512, classical_final_output),
        )

    def forward(self, x1):
        x1 = self.features(x1)
        # x2 = self.features(x2)
        return x1

class SigNet__Half_Half_Single(nn.Module):
    def __init__(self, classical_final_output=128):
        super().__init__()
        self.features = nn.Sequential(
            # Input: [1, 155, 220]
            nn.Conv2d(1, 24, kernel_size=11, stride=4),  # [96, 37, 53]
            nn.ReLU(),
            nn.LocalResponseNorm(size=5, k=2, alpha=1e-4, beta=0.75),
            nn.MaxPool2d(kernel_size=3, stride=2),  # [96, 18, 26]
            nn.Conv2d(24, 64, kernel_size=5, padding=2),  # [256, 18, 26]
            nn.LocalResponseNorm(size=5, k=2, alpha=1e-4, beta=0.75),
            nn.MaxPool2d(kernel_size=3, stride=2),  # [256, 8, 12]
            nn.Dropout2d(p=0.3),
            nn.Conv2d(64, 96, kernel_size=3, padding=1),  # [384, 8, 12]
            nn.Conv2d(96, 64, kernel_size=3, padding=1),  # [256, 8, 12]
            nn.MaxPool2d(kernel_size=3, stride=2),  # [256, 3, 5]
            nn.Dropout2d(p=0.3),
            nn.Flatten(),  # 256*3*5 = 3,840
            nn.Linear(960, 256),
            nn.Dropout(p=0.5),  # Note: Dropout2d → Dropout for Linear layers
            nn.Linear(256, classical_final_output),
        )

    def forward(self, x1):
        x1 = self.features(x1)
        # x2 = self.features(x2)
        return x1

class SigNet(nn.Module):
    def __init__(self):
        super().__init__()
        self.features = nn.Sequential(
            # Input: [1, 155, 220]
            nn.Conv2d(1, 96, kernel_size=11, stride=4),  # [96, 37, 53]
            nn.ReLU(),
            nn.LocalResponseNorm(size=5, k=2, alpha=1e-4, beta=0.75),
            nn.MaxPool2d(kernel_size=3, stride=2),  # [96, 18, 26]
            nn.Conv2d(96, 256, kernel_size=5, padding=2),  # [256, 18, 26]
            nn.LocalResponseNorm(size=5, k=2, alpha=1e-4, beta=0.75),
            nn.MaxPool2d(kernel_size=3, stride=2),  # [256, 8, 12]
            nn.Dropout2d(p=0.3),
            nn.Conv2d(256, 384, kernel_size=3, padding=1),  # [384, 8, 12]
            nn.Conv2d(384, 256, kernel_size=3, padding=1),  # [256, 8, 12]
            nn.MaxPool2d(kernel_size=3, stride=2),  # [256, 3, 5]
            nn.Dropout2d(p=0.3),
            nn.Flatten(),  # 256*3*5 = 3,840
            nn.Linear(3840, 1024),
            nn.Dropout(p=0.5),  # Note: Dropout2d → Dropout for Linear layers
            nn.Linear(1024, 128),
        )

    def forward(self, x1, x2):
        x1 = self.features(x1)
        x2 = self.features(x2)
        return x1, x2

class SigNet__Single(nn.Module):
    def __init__(self):
        super().__init__()
        self.features = nn.Sequential(
            # Input: [1, 155, 220]
            nn.Conv2d(1, 96, kernel_size=11, stride=4),  # [96, 37, 53]
            nn.ReLU(),
            nn.LocalResponseNorm(size=5, k=2, alpha=1e-4, beta=0.75),
            nn.MaxPool2d(kernel_size=3, stride=2),  # [96, 18, 26]
            nn.Conv2d(96, 256, kernel_size=5, padding=2),  # [256, 18, 26]
            nn.LocalResponseNorm(size=5, k=2, alpha=1e-4, beta=0.75),
            nn.MaxPool2d(kernel_size=3, stride=2),  # [256, 8, 12]
            nn.Dropout2d(p=0.3),
            nn.Conv2d(256, 384, kernel_size=3, padding=1),  # [384, 8, 12]
            nn.Conv2d(384, 256, kernel_size=3, padding=1),  # [256, 8, 12]
            nn.MaxPool2d(kernel_size=3, stride=2),  # [256, 3, 5]
            nn.Dropout2d(p=0.3),
            nn.Flatten(),  # 256*3*5 = 3,840
            nn.Linear(3840, 1024),
            nn.Dropout(p=0.5),  # Note: Dropout2d → Dropout for Linear layers
            nn.Linear(1024, 128),
        )

    def forward(self, x1):
        x1 = self.features(x1)
        # x2 = self.features(x2)
        return x1

class SigNet2_FullClassical(nn.Module):
    '''
    Reference Keras: https://github.com/sounakdey/SigNet/blob/master/SigNet_v1.py
    '''
    def __init__(self):
        super().__init__()
        # self.quantum_weights = {"weights": (4, n_qubits, 3)}
        # self.qlayer = qml.qnn.TorchLayer(Naive_QuantumHybrid, self.quantum_weights)

        self.features = nn.Sequential(
            #input size = [155, 220, 1]
            nn.Conv2d(1, 48, 11), # size = [145,210]
            nn.ReLU(),
            nn.LocalResponseNorm(size=5, k=2, alpha=1e-4, beta=0.75),
            nn.MaxPool2d(2, stride=2), # size = [72, 105]
            nn.Conv2d(48, 64, 5, padding=2, padding_mode='zeros'), # size = [72, 105]
            nn.LocalResponseNorm(size=5, k=2, alpha=1e-4, beta=0.75),
            nn.MaxPool2d(2, stride=2), # size = [36, 52]
            nn.Dropout2d(p=0.3),
            nn.Conv2d(64, 128, 3, stride=1, padding=1, padding_mode='zeros'),
            nn.Conv2d(128, 64, 3, stride=1, padding=1, padding_mode='zeros'),
            nn.MaxPool2d(2, stride=2), # size = [18, 26]
            nn.Dropout2d(p=0.3),
            nn.Flatten(1, -1), # 18*26*256
            nn.Linear(18*26*64, 1024),
            nn.Dropout2d(p=0.5),
            nn.Linear(1024, 128),
            # nn.Dropout2d(p=0.5),
            # nn.Linear(128, 8)
            # nn.Dropout2d(p=0.5),
            # self.qlayer,
            # nn.Linear(8, 16),
            # nn.Dropout2d(p=0.5),
            # nn.Linear(16, 32)
        )

        # TODO: init bias = 0

    def forward(self, x1, x2):
        x1 = self.features(x1)
        x2 = self.features(x2)
        return x1, x2

class SigNet2_FullClassical__Single(nn.Module):
    '''
    Reference Keras: https://github.com/sounakdey/SigNet/blob/master/SigNet_v1.py
    '''
    def __init__(self):
        super().__init__()
        # self.quantum_weights = {"weights": (4, n_qubits, 3)}
        # self.qlayer = qml.qnn.TorchLayer(Naive_QuantumHybrid, self.quantum_weights)

        self.features = nn.Sequential(
            #input size = [155, 220, 1]
            nn.Conv2d(1, 48, 11), # size = [145,210]
            nn.ReLU(),
            nn.LocalResponseNorm(size=5, k=2, alpha=1e-4, beta=0.75),
            nn.MaxPool2d(2, stride=2), # size = [72, 105]
            nn.Conv2d(48, 64, 5, padding=2, padding_mode='zeros'), # size = [72, 105]
            nn.LocalResponseNorm(size=5, k=2, alpha=1e-4, beta=0.75),
            nn.MaxPool2d(2, stride=2), # size = [36, 52]
            nn.Dropout2d(p=0.3),
            nn.Conv2d(64, 128, 3, stride=1, padding=1, padding_mode='zeros'),
            nn.Conv2d(128, 64, 3, stride=1, padding=1, padding_mode='zeros'),
            nn.MaxPool2d(2, stride=2), # size = [18, 26]
            nn.Dropout2d(p=0.3),
            nn.Flatten(1, -1), # 18*26*256
            nn.Linear(18*26*64, 1024),
            nn.Dropout2d(p=0.5),
            nn.Linear(1024, 128),
            # nn.Dropout2d(p=0.5),
            # nn.Linear(128, 8)
            # nn.Dropout2d(p=0.5),
            # self.qlayer,
            # nn.Linear(8, 16),
            # nn.Dropout2d(p=0.5),
            # nn.Linear(16, 32)
        )

        # TODO: init bias = 0

    def forward(self, x1):
        x1 = self.features(x1)
        # x2 = self.features(x2)
        return x1


import os
from PIL import ImageOps, Image

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torchvision import transforms
from torch.utils.data import Dataset, DataLoader, IterableDataset

from sklearn.model_selection import train_test_split
from sklearn.utils import shuffle

import pandas as pd

import random
import csv
import itertools
import pickle

import matplotlib.pyplot as plt
import cv2

import numpy as np
import tqdm

import pennylane as qml

from sklearn import metrics
from sklearn.metrics import classification_report

import helper_funcs
from helper_funcs import n_qubits_quantum_single_channel, n_qubits_single_channel


def qml_RZZ(swap_angle, wires):
    qml.CNOT(wires=wires)
    qml.RZ(swap_angle, wires=wires[1])
    qml.CNOT(wires=wires)

# @qml.qnode(dev, interface="torch")
def I_Block_Vanilla__1to1(n_qubits):
    dev = qml.device("default.qubit", n_qubits)

    def circuit(q_input_features_input1, q_input_features_input2, q_variational_swaps, pre_hadamard=False, final_hadamard=False):
        # encoding the outputs from the
        for i in range(n_qubits_single_channel):
            qml.RY(q_input_features_input1[i]*2*np.pi, wires=i)
        for i in range(n_qubits_single_channel, 2*n_qubits_single_channel):
            qml.RY(q_input_features_input2[i - n_qubits_single_channel]*2*np.pi, wires=i)

        # CNOT blocks
        for i in range(n_qubits_single_channel):
            qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])
        for i in range(n_qubits_single_channel):
            qml.CNOT(wires=[n_qubits_single_channel+i, n_qubits_single_channel+((i+1)%n_qubits_single_channel)])

        # parametric interference
        for i in range(n_qubits_single_channel):
            # decomposing the RZZ gate; not available in PennyLane as of 27 May 2025
            # qml.CNOT(wires=[i, i + n_qubits_single_channel])
            # qml.RZ(q_variational_swaps[i]*2*np.pi, wires=i + n_qubits_single_channel)
            # qml.CNOT(wires=[i, i + n_qubits_single_channel])

            qml_RZZ(q_variational_swaps[i]*2*np.pi, wires=[i, i + n_qubits_single_channel])

        # if final_hadamard:
        #     for i in range(n_qubits):
        #         qml.Hadamard(wires=i)

        # CNOT blocks
        for i in range(n_qubits_single_channel):
            qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])
        for i in range(n_qubits_single_channel):
            qml.CNOT(wires=[n_qubits_single_channel+i, n_qubits_single_channel+((i+1)%n_qubits_single_channel)])

        if final_hadamard:
            for i in range(n_qubits):
                qml.Hadamard(wires=i)

        expvals = [qml.expval(qml.PauliZ(i)) for i in range(n_qubits)]
        return tuple(expvals[:n_qubits_single_channel]), tuple(expvals[n_qubits_single_channel:])
    
    return qml.QNode(circuit, dev, interface="torch")

# @qml.qnode(dev, interface="torch")
def I_Block_AmplitudeEnc__variational_swap_cyclic_ansatz__w_per_pre_post_VQC(n_qubits):
    dev = qml.device("default.qubit", n_qubits)
    def circuit(inputs, q_weights_flat, pre_per_channel_VQC_weights, post_per_channel_VQC_weights):
        q_input_features_input1 = inputs[:, :2**n_qubits_single_channel]
        q_input_features_input2 = inputs[:, 2**n_qubits_single_channel:]

        # some hyper-params
        pre_per_channel_VQC = True
        post_per_channel_VQC = True
        pre_per_channel_VQC_depth = 2
        post_per_channel_VQC_depth = 2
        q_depth = 3
        pre_hadamard = False
        final_hadamard = False

        # # encoding the outputs from the
        # for i in range(n_qubits_single_channel):
        #     qml.RY(q_input_features_input1[:, i]*np.pi, wires=i)
        # for i in range(n_qubits_single_channel, 2*n_qubits_single_channel):
        #     qml.RY(q_input_features_input2[:, i - n_qubits_single_channel]*np.pi, wires=i)

        # encoding using Amplitude Embedding
        q_input_features_input1 = F.normalize(q_input_features_input1, p=2, dim=1)
        q_input_features_input2 = F.normalize(q_input_features_input2, p=2, dim=1)
        # state_vector = torch.cat((q_input_features_input1, q_input_features_input2), dim=1)

        qml.AmplitudeEmbedding(features=q_input_features_input1, wires=[i for i in range(n_qubits_single_channel)], normalize=True)
        # qml.AmplitudeEmbedding(features=q_input_features_input2, wires=[i for i in range(n_qubits_single_channel, 2*n_qubits_single_channel, 1)], normalize=True)

        # # CNOT blocks
        # for i in range(n_qubits_single_channel):
        #     qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])
        # for i in range(n_qubits_single_channel):
        #     qml.CNOT(wires=[n_qubits_single_channel+i, n_qubits_single_channel+((i+1)%n_qubits_single_channel)])

        # per-channel pre VQC
        if pre_per_channel_VQC and (pre_per_channel_VQC_weights is not None):
            # pre_per_channel_VQC_weights = pre_per_channel_VQC_weights.reshape(pre_per_channel_VQC_depth, n_qubits_single_channel, 3)

            for layer_idx in range(pre_per_channel_VQC_depth):
                for qubit_idx in range(n_qubits_single_channel):
                    qml.U3(*pre_per_channel_VQC_weights[layer_idx, qubit_idx, :], wires=qubit_idx)
                    qml.U3(*pre_per_channel_VQC_weights[layer_idx, qubit_idx, :], wires=n_qubits_single_channel+qubit_idx)

                # CNOT blocks
                for i in range(n_qubits_single_channel):
                    qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])
                for i in range(n_qubits_single_channel):
                    qml.CNOT(wires=[n_qubits_single_channel+i, n_qubits_single_channel+((i+1)%n_qubits_single_channel)])

        # # parametric interference - from one wire in upper branch to all in lower branch, all different
        # for i in range(n_qubits_single_channel):
        #     for j in range(n_qubits_single_channel):
        #         qml_RZZ(q_variational_swaps[i+j]*2*np.pi, wires=[i, j + n_qubits_single_channel])

        # parametric interference - VQC for weight mixing
        # adding the cyclic CP gates
        wire_selector = lambda x : (x//2) + (x%2)*n_qubits_single_channel
        for angle_idx, angle in enumerate(q_weights_flat):
            control = wire_selector(angle_idx)
            target = wire_selector((angle_idx+1)%n_qubits)

            qml.ctrl(qml.RY, control=control)(angle, wires=target)

        # # add the circuit for the VQC (hardware-efficient ansatz)
        # q_weights = q_weights_flat.reshape(q_depth, n_qubits, 3)
        # for layer_idx in range(q_depth):
        #     for qubit_idx in range(n_qubits):
        #         qml.U3(q_weights[layer_idx, qubit_idx, :], wires=qubit_idx)

        #     # hardware-efficient ensatz
        #     for qubit_idx in range(0, n_qubits, 2):
        #         qml.CNOT(wires=[qubit_idx, qubit_idx + 1])
        #     for qubit_idx in range(1, n_qubits, 2):
        #         qml.CNOT(wires=[qubit_idx, (qubit_idx + 1) % n_qubits])

        # if final_hadamard:
        #     for i in range(n_qubits):
        #         qml.Hadamard(wires=i)

        # # CNOT blocks
        # for i in range(n_qubits_single_channel):
        #     qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])
        # for i in range(n_qubits_single_channel):
        #     qml.CNOT(wires=[n_qubits_single_channel+i, n_qubits_single_channel+((i+1)%n_qubits_single_channel)])

        # per-channel post VQC
        if post_per_channel_VQC and (post_per_channel_VQC_weights is not None):
            # post_per_channel_VQC_weights = post_per_channel_VQC_weights.reshape(post_per_channel_VQC_depth, n_qubits_single_channel, 3)

            for layer_idx in range(post_per_channel_VQC_depth):
                for qubit_idx in range(n_qubits_single_channel):
                    qml.U3(*post_per_channel_VQC_weights[layer_idx, qubit_idx, :], wires=qubit_idx)
                    qml.U3(*post_per_channel_VQC_weights[layer_idx, qubit_idx, :], wires=n_qubits_single_channel+qubit_idx)

                # CNOT blocks
                for i in range(n_qubits_single_channel):
                    qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])
                for i in range(n_qubits_single_channel):
                    qml.CNOT(wires=[n_qubits_single_channel+i, n_qubits_single_channel+((i+1)%n_qubits_single_channel)])

        if final_hadamard:
            for i in range(n_qubits):
                qml.Hadamard(wires=i)

        expvals =  [qml.expval(qml.PauliZ(i)) for i in range(n_qubits)]
        # return tuple(expvals)
        return expvals[:n_qubits_single_channel], expvals[n_qubits_single_channel:]
    
    return qml.QNode(circuit, dev, interface="torch")

# @qml.qnode(dev, interface="torch")
def circuit__1__AngleEmbedding(n_qubits_main=512, s_dagger=False):
    n_qubits = n_qubits_quantum_single_channel(n_qubits_main)
    dev = qml.device("default.qubit", n_qubits)
    kappa = 1
    random_inds = sorted(np.random.choice(n_qubits - 1, size=kappa, replace=False))

    def circuit(inputs, theta_RY_switch, vqc_weights):
        # vqc_weights shape: (num_blocks, num_layers_per_block, num_wires, 3)
        switch = n_qubits - 1
        if s_dagger:
            qml.adjoint(qml.S(wires=switch))

        num_vqc_blocks = vqc_weights.shape[0]
        num_layers_per_block = vqc_weights.shape[1]
        num_wires_vqc = vqc_weights.shape[2]
        # print(num_wires_vqc, switch)
        assert num_wires_vqc == switch

        qml.Hadamard(wires=switch)
        # print("shape of input for circuit__1:", inputs.shape)

        # assuming that the input is normalized in the specific way
        # qml.AmplitudeEmbedding(features=inputs, wires=[w for w in range(n_qubits)], normalize=False)
        qml.ctrl(qml.AngleEmbedding, switch, control_values=(0))(features=inputs[:, :switch], wires=[w for w in range(switch)], rotation='Y')
        qml.ctrl(qml.AngleEmbedding, switch, control_values=(1))(features=inputs[:, switch:], wires=[w for w in range(switch)], rotation='Y')
        
        # qml.MottonenStatePreparation(state_vector=inputs, wires=[w for w in range(n_qubits)])

        # if (pre_vqc is True) and (pre_vqc_weights is not None):
        if True:
            for b in range(num_vqc_blocks):
                # qml.StronglyEntanglingLayers(weights=vqc_weights[b], wires=range(switch))

                for l in range(num_layers_per_block):
                    for q in range(num_wires_vqc):
                        qml.U3(*vqc_weights[b, l, q, :], wires=q)

                # # cyclic CNOT on 8
                # for i in range(n_qubits_single_channel):
                #     qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])

                # hardware-efficient CNOT on 8
                for i in range(0, num_wires_vqc, 2):
                    qml.CNOT(wires=[i, (i+1)%num_wires_vqc])
                for i in range(1, num_wires_vqc, 2):
                    qml.CNOT(wires=[i, (i+1)%num_wires_vqc])

                # qml.ctrl(qml.RX, (b % switch), control_values=(1))(theta_RY_switch[b], wires=switch)
                qml.ctrl(qml.U3, random_inds, control_values=(1 for _ in range(kappa)))(*theta_RY_switch[b], wires=switch)

        # final overlap
        qml.Hadamard(wires=switch)
        return qml.expval(qml.PauliZ(wires=switch))
    
    return qml.QNode(circuit, dev, interface="torch")

# @qml.qnode(dev, interface="torch")
def type2_classical_quantum__single_branch(num_single_channel=128):
    n_qubits = n_qubits_quantum_single_channel(num_single_channel)
    dev = qml.device("default.qubit", n_qubits)

    def circuit(inputs, vqc_weights):
        # vqc_weights shape: (num_blocks, num_wires, 3)
        num_vqc_blocks = vqc_weights.shape[0]
        num_wires_vqc = vqc_weights.shape[1]

        # assuming that the input is normalized in the specific way
        qml.AmplitudeEmbedding(features=inputs, wires=[w for w in range(n_qubits)], normalize=True)

        if (vqc_weights is not None):
        # if True:
            for l in range(num_vqc_blocks):
                for q in range(num_wires_vqc):
                    qml.U3(*vqc_weights[l, q, :], wires=q)

                # # cyclic CNOT on 8
                # for i in range(n_qubits_single_channel):
                #     qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])

                # hardware-efficient CNOT on 8
                for i in range(0, num_wires_vqc, 2):
                    qml.CNOT(wires=[i, (i+1)%num_wires_vqc])
                for i in range(1, num_wires_vqc, 2):
                    qml.CNOT(wires=[i, (i+1)%num_wires_vqc])
        
        return [qml.expval(qml.PauliZ(wires=i)) for i in range(num_wires_vqc)]
    
    return qml.QNode(circuit, dev, interface="torch")

# @qml.qnode(dev, interface="torch")
def I_Block_Vanilla__variational_swap_cyclic_ansatz__w_per_pre_post_VQC(n_qubits):
    dev = qml.device("default.qubit", n_qubits)

    def circuit(inputs, q_weights_flat, pre_per_channel_VQC_weights, post_per_channel_VQC_weights):
        q_input_features_input1 = inputs[:, :n_qubits_single_channel]
        q_input_features_input2 = inputs[:, n_qubits_single_channel:]

        # some hyper-params
        pre_per_channel_VQC = True
        post_per_channel_VQC = True
        pre_per_channel_VQC_depth = 2
        post_per_channel_VQC_depth = 2
        q_depth = 3
        pre_hadamard = False
        final_hadamard = False

        # encoding the outputs from the
        for i in range(n_qubits_single_channel):
            qml.RY(q_input_features_input1[:, i]*np.pi, wires=i)
        for i in range(n_qubits_single_channel, 2*n_qubits_single_channel):
            qml.RY(q_input_features_input2[:, i - n_qubits_single_channel]*np.pi, wires=i)

        # CNOT blocks
        for i in range(n_qubits_single_channel):
            qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])
        for i in range(n_qubits_single_channel):
            qml.CNOT(wires=[n_qubits_single_channel+i, n_qubits_single_channel+((i+1)%n_qubits_single_channel)])

        # per-channel pre VQC
        if pre_per_channel_VQC and (pre_per_channel_VQC_weights is not None):
            # pre_per_channel_VQC_weights = pre_per_channel_VQC_weights.reshape(pre_per_channel_VQC_depth, n_qubits_single_channel, 3)

            for layer_idx in range(pre_per_channel_VQC_depth):
                for qubit_idx in range(n_qubits_single_channel):
                    qml.U3(*pre_per_channel_VQC_weights[layer_idx, qubit_idx, :], wires=qubit_idx)
                    qml.U3(*pre_per_channel_VQC_weights[layer_idx, qubit_idx, :], wires=n_qubits_single_channel+qubit_idx)

                # CNOT blocks
                for i in range(n_qubits_single_channel):
                    qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])
                for i in range(n_qubits_single_channel):
                    qml.CNOT(wires=[n_qubits_single_channel+i, n_qubits_single_channel+((i+1)%n_qubits_single_channel)])

        # # parametric interference - from one wire in upper branch to all in lower branch, all different
        # for i in range(n_qubits_single_channel):
        #     for j in range(n_qubits_single_channel):
        #         qml_RZZ(q_variational_swaps[i+j]*2*np.pi, wires=[i, j + n_qubits_single_channel])

        # parametric interference - VQC for weight mixing
        # adding the cyclic CP gates
        wire_selector = lambda x : (x//2) + (x%2)*n_qubits_single_channel
        for angle_idx, angle in enumerate(q_weights_flat):
            control = wire_selector(angle_idx)
            target = wire_selector((angle_idx+1)%n_qubits)

            qml.ctrl(qml.RY, control=control)(angle, wires=target)

        # # add the circuit for the VQC (hardware-efficient ansatz)
        # q_weights = q_weights_flat.reshape(q_depth, n_qubits, 3)
        # for layer_idx in range(q_depth):
        #     for qubit_idx in range(n_qubits):
        #         qml.U3(q_weights[layer_idx, qubit_idx, :], wires=qubit_idx)

        #     # hardware-efficient ensatz
        #     for qubit_idx in range(0, n_qubits, 2):
        #         qml.CNOT(wires=[qubit_idx, qubit_idx + 1])
        #     for qubit_idx in range(1, n_qubits, 2):
        #         qml.CNOT(wires=[qubit_idx, (qubit_idx + 1) % n_qubits])

        # if final_hadamard:
        #     for i in range(n_qubits):
        #         qml.Hadamard(wires=i)

        # CNOT blocks
        for i in range(n_qubits_single_channel):
            qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])
        for i in range(n_qubits_single_channel):
            qml.CNOT(wires=[n_qubits_single_channel+i, n_qubits_single_channel+((i+1)%n_qubits_single_channel)])

        # per-channel post VQC
        if post_per_channel_VQC and (post_per_channel_VQC_weights is not None):
            # post_per_channel_VQC_weights = post_per_channel_VQC_weights.reshape(post_per_channel_VQC_depth, n_qubits_single_channel, 3)

            for layer_idx in range(post_per_channel_VQC_depth):
                for qubit_idx in range(n_qubits_single_channel):
                    qml.U3(*post_per_channel_VQC_weights[layer_idx, qubit_idx, :], wires=qubit_idx)
                    qml.U3(*post_per_channel_VQC_weights[layer_idx, qubit_idx, :], wires=n_qubits_single_channel+qubit_idx)

                # CNOT blocks
                for i in range(n_qubits_single_channel):
                    qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])
                for i in range(n_qubits_single_channel):
                    qml.CNOT(wires=[n_qubits_single_channel+i, n_qubits_single_channel+((i+1)%n_qubits_single_channel)])

        if final_hadamard:
            for i in range(n_qubits):
                qml.Hadamard(wires=i)

        expvals =  [qml.expval(qml.PauliZ(i)) for i in range(n_qubits)]
        # return tuple(expvals)
        return expvals[:n_qubits_single_channel], expvals[n_qubits_single_channel:]
    
    return qml.QNode(circuit, dev, interface="torch")

# @qml.qnode(dev, interface="torch")
def AmpEmb__Vanilla__pre_post_VQC__switch_ry(n_qubits):
    dev = qml.device("default.qubit", n_qubits)

    def circuit(inputs, theta_RY_switch, pre_vqc_weights, post_vqc_weights):
        num_pre_vqc_layers = pre_vqc_weights.shape[0]
        num_post_vqc_layers = post_vqc_weights.shape[0]
        pre_vqc = True
        post_vqc = True

        # assuming that the input is normalized in the specific way
        qml.AmplitudeEmbedding(features=inputs, wires=[w for w in range(n_qubits)], normalize=False)

        # RY on 9th qubit for weight sharing
        # qml.RY(phi=np.pi/8, wires=8)
        qml.RY(theta_RY_switch, wires=n_qubits-1)

        # pre_vqc_weights -> (num_layers, num_qubits, 3)
        # hardware-efficient entangling for 8
        if (pre_vqc is True) and (pre_vqc_weights is not None):
            for l in range(num_pre_vqc_layers):
                for q in range(n_qubits_single_channel):
                    qml.U3(*pre_vqc_weights[l, q, :], wires=q)

                # # cyclic CNOT on 8
                # for i in range(n_qubits_single_channel):
                #     qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])

                # hardware-efficient CNOT on 8
                for i in range(0, n_qubits_single_channel, 2):
                    qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])
                for i in range(1, n_qubits_single_channel, 2):
                    qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])

        # post_vqc_weights -> (num_layers, num_qubits, 3)
        # hardware-efficient entangling for 8
        if (post_vqc is True) and (post_vqc_weights is not None):
            for l in range(num_post_vqc_layers):
                for q in range(n_qubits_single_channel):
                    qml.U3(*post_vqc_weights[l, q, :], wires=q)

                # # cyclic CNOT on 8
                # for i in range(n_qubits_single_channel):
                #     qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])

                # hardware-efficient CNOT on 8
                for i in range(0, n_qubits_single_channel, 2):
                    qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])
                for i in range(1, n_qubits_single_channel, 2):
                    qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])

        # final overlap
        qml.Hadamard(wires=n_qubits-1)
        return qml.expval(qml.PauliZ(wires=n_qubits-1))
    
    return qml.QNode(circuit, dev, interface="torch")


# @qml.qnode(dev, interface="torch")
def AmpEmb__pre__c_post_VQC__switch_ry(n_qubits):
    dev = qml.device("default.qubit", n_qubits)

    def circuit(inputs,
                pre_vqc_weights_A, post_vqc_weights_A,
                pre_vqc_weights_B, post_vqc_weights_B,
                pre_vqc_weights_C, post_vqc_weights_C,
                pre_vqc_weights_D, post_vqc_weights_D):
        num_pre_vqc_layers = pre_vqc_weights_A.shape[0]
        num_post_vqc_layers = post_vqc_weights_A.shape[0]
        pre_vqc = True
        post_vqc = True

        theta_RY_switch_1 = np.pi/2
        theta_RY_switch_2 = np.pi/3
        theta_RY_switch_3 = np.pi/4

        # assuming that the input is normalized in the specific way
        qml.AmplitudeEmbedding(features=inputs, wires=[w for w in range(n_qubits)], normalize=False)

        def pre_VQC_A():
            # pre_vqc_weights -> (num_layers, num_qubits, 3)
            # hardware-efficient entangling for 8
            if (pre_vqc is True) and (pre_vqc_weights_A is not None):
                for l in range(num_pre_vqc_layers):
                    for q in range(n_qubits_single_channel):
                        qml.U3(*pre_vqc_weights_A[l, q, :], wires=q)

                    # # cyclic CNOT on 8
                    # for i in range(n_qubits_single_channel):
                    #     qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])

                    # hardware-efficient CNOT on 8
                    for i in range(0, n_qubits_single_channel, 2):
                        qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])
                    for i in range(1, n_qubits_single_channel, 2):
                        qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])
        def post_VQC_A():
            # post_vqc_weights -> (num_layers, num_qubits, 3)
            # hardware-efficient entangling for 8
            if (post_vqc is True) and (post_vqc_weights_A is not None):
                for l in range(num_post_vqc_layers):
                    for q in range(n_qubits_single_channel):
                        qml.U3(*post_vqc_weights_A[l, q, :], wires=q)

                    # # cyclic CNOT on 8
                    # for i in range(n_qubits_single_channel):
                    #     qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])

                    # hardware-efficient CNOT on 8
                    for i in range(0, n_qubits_single_channel, 2):
                        qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])
                    for i in range(1, n_qubits_single_channel, 2):
                        qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])

        def pre_VQC_B():
            # pre_vqc_weights -> (num_layers, num_qubits, 3)
            # hardware-efficient entangling for 8
            if (pre_vqc is True) and (pre_vqc_weights_B is not None):
                for l in range(num_pre_vqc_layers):
                    for q in range(n_qubits_single_channel):
                        qml.U3(*pre_vqc_weights_B[l, q, :], wires=q)

                    # # cyclic CNOT on 8
                    # for i in range(n_qubits_single_channel):
                    #     qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])

                    # hardware-efficient CNOT on 8
                    for i in range(0, n_qubits_single_channel, 2):
                        qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])
                    for i in range(1, n_qubits_single_channel, 2):
                        qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])
        def post_VQC_B():
            # post_vqc_weights -> (num_layers, num_qubits, 3)
            # hardware-efficient entangling for 8
            if (post_vqc is True) and (post_vqc_weights_B is not None):
                for l in range(num_post_vqc_layers):
                    for q in range(n_qubits_single_channel):
                        qml.U3(*post_vqc_weights_B[l, q, :], wires=q)

                    # # cyclic CNOT on 8
                    # for i in range(n_qubits_single_channel):
                    #     qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])

                    # hardware-efficient CNOT on 8
                    for i in range(0, n_qubits_single_channel, 2):
                        qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])
                    for i in range(1, n_qubits_single_channel, 2):
                        qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])

        def pre_VQC_C():
            # pre_vqc_weights -> (num_layers, num_qubits, 3)
            # hardware-efficient entangling for 8
            if (pre_vqc is True) and (pre_vqc_weights_C is not None):
                for l in range(num_pre_vqc_layers):
                    for q in range(n_qubits_single_channel):
                        qml.U3(*pre_vqc_weights_C[l, q, :], wires=q)

                    # # cyclic CNOT on 8
                    # for i in range(n_qubits_single_channel):
                    #     qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])

                    # hardware-efficient CNOT on 8
                    for i in range(0, n_qubits_single_channel, 2):
                        qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])
                    for i in range(1, n_qubits_single_channel, 2):
                        qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])
        def post_VQC_C():
            # post_vqc_weights -> (num_layers, num_qubits, 3)
            # hardware-efficient entangling for 8
            if (post_vqc is True) and (post_vqc_weights_C is not None):
                for l in range(num_post_vqc_layers):
                    for q in range(n_qubits_single_channel):
                        qml.U3(*post_vqc_weights_C[l, q, :], wires=q)

                    # # cyclic CNOT on 8
                    # for i in range(n_qubits_single_channel):
                    #     qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])

                    # hardware-efficient CNOT on 8
                    for i in range(0, n_qubits_single_channel, 2):
                        qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])
                    for i in range(1, n_qubits_single_channel, 2):
                        qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])

        def pre_VQC_D():
            # pre_vqc_weights -> (num_layers, num_qubits, 3)
            # hardware-efficient entangling for 8
            if (pre_vqc is True) and (pre_vqc_weights_D is not None):
                for l in range(num_pre_vqc_layers):
                    for q in range(n_qubits_single_channel):
                        qml.U3(*pre_vqc_weights_D[l, q, :], wires=q)

                    # # cyclic CNOT on 8
                    # for i in range(n_qubits_single_channel):
                    #     qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])

                    # hardware-efficient CNOT on 8
                    for i in range(0, n_qubits_single_channel, 2):
                        qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])
                    for i in range(1, n_qubits_single_channel, 2):
                        qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])
        def post_VQC_D():
            # post_vqc_weights -> (num_layers, num_qubits, 3)
            # hardware-efficient entangling for 8
            if (post_vqc is True) and (post_vqc_weights_D is not None):
                for l in range(num_post_vqc_layers):
                    for q in range(n_qubits_single_channel):
                        qml.U3(*post_vqc_weights_D[l, q, :], wires=q)

                    # # cyclic CNOT on 8
                    # for i in range(n_qubits_single_channel):
                    #     qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])

                    # hardware-efficient CNOT on 8
                    for i in range(0, n_qubits_single_channel, 2):
                        qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])
                    for i in range(1, n_qubits_single_channel, 2):
                        qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])

        qml.X(wires=n_qubits-1)
        qml.ctrl(pre_VQC_A, control=n_qubits-1)()
        qml.X(wires=n_qubits-1)
        qml.ctrl(post_VQC_A, control=n_qubits-1)()

        # RY on 9th qubit for weight sharing
        # qml.RY(phi=np.pi/8, wires=8)
        qml.RY(theta_RY_switch_1, wires=n_qubits-1)

        qml.X(wires=n_qubits-1)
        qml.ctrl(pre_VQC_B, control=n_qubits-1)()
        qml.X(wires=n_qubits-1)
        qml.ctrl(post_VQC_B, control=n_qubits-1)()

        qml.RY(theta_RY_switch_2, wires=n_qubits-1)

        qml.X(wires=n_qubits-1)
        qml.ctrl(pre_VQC_C, control=n_qubits-1)()
        qml.X(wires=n_qubits-1)
        qml.ctrl(post_VQC_C, control=n_qubits-1)()

        qml.RY(theta_RY_switch_3, wires=n_qubits-1)

        qml.X(wires=n_qubits-1)
        qml.ctrl(pre_VQC_D, control=n_qubits-1)()
        qml.X(wires=n_qubits-1)
        qml.ctrl(post_VQC_D, control=n_qubits-1)()

        # final overlap
        qml.Hadamard(wires=n_qubits-1)
        return qml.expval(qml.PauliZ(wires=n_qubits-1))
    
    return qml.QNode(circuit, dev, interface="torch")

# @qml.qnode(dev, interface="torch")
def circuit__1(bool_vals: list, n_blocks=3, kappa=4, length_total_channels=256):
    type1_local_cost = False
    type2_local_cost = False
    weight_interaction_cost = False
    weight_interaction_cost__relaxed = False
    global_cost__layer_wise_type2 = False

    costs = [type1_local_cost, type2_local_cost, weight_interaction_cost, weight_interaction_cost__relaxed, global_cost__layer_wise_type2]
    for b in bool_vals:
        costs[b] = True

    s_dagger = False

    n_qubits_wo_blockwise = n_qubits_quantum_single_channel(length_total_channels)
    n_qubits = n_qubits_wo_blockwise
    if (weight_interaction_cost is True) or (weight_interaction_cost__relaxed is True):
        n_qubits = n_qubits_wo_blockwise + n_blocks

    dev = qml.device("default.qubit", n_qubits)
    # random_inds = sorted(np.random.choice(n_qubits - 1, size=kappa, replace=False))

    control_vals = [1 for _ in range(kappa)]
    control_vals_mixed = True
    if control_vals_mixed is True:
        control_vals = [1] * (kappa // 2) + [0] * (kappa - (kappa // 2))
        random.shuffle(control_vals)

    all_layers_random_ctlrs = True
    random_inds_list = []

    def circuit(inputs, theta_RY_switch, vqc_weights):
        # vqc_weights shape: (num_blocks, num_layers_per_block, num_wires, 3)
        switch = n_qubits_wo_blockwise - 1
        global control_vals

        num_vqc_blocks = vqc_weights.shape[0]
        num_layers_per_block = vqc_weights.shape[1]
        num_wires_vqc = vqc_weights.shape[2]
        # print(num_wires_vqc, switch)
        assert num_wires_vqc == switch

        # qml.Hadamard(wires=switch)
        # print("shape of input for circuit__1:", inputs)

        # assuming that the input is normalized in the specific way
        qml.AmplitudeEmbedding(features=inputs, wires=[w for w in range(n_qubits_wo_blockwise)], normalize=False)
        # qml.ctrl(qml.AmplitudeEmbedding, switch, control_values=(0))(features=inputs[:, :128], wires=[w for w in range(switch)], normalize=True)
        # qml.ctrl(qml.AmplitudeEmbedding, switch, control_values=(1))(features=inputs[:, 128:], wires=[w for w in range(switch)], normalize=True)

        if s_dagger:
            qml.adjoint(qml.S(wires=switch))

        if (weight_interaction_cost is True) or (weight_interaction_cost__relaxed is True):
            for _ in range(n_qubits_wo_blockwise, n_qubits, 1):
                qml.Hadamard(wires=_)
        
        # qml.MottonenStatePreparation(state_vector=inputs, wires=[w for w in range(n_qubits)])
        # print("qml state:", qml.state())

        # if (pre_vqc is True) and (pre_vqc_weights is not None):
        if True:
            for b in range(num_vqc_blocks):
                # qml.StronglyEntanglingLayers(weights=vqc_weights[b], wires=range(switch))

                for l in range(num_layers_per_block):
                    for q in range(num_wires_vqc):
                        qml.U3(*vqc_weights[b, l, q, :], wires=q)

                # # cyclic CNOT on 8
                # for i in range(n_qubits_single_channel):
                #     qml.CNOT(wires=[i, (i+1)%n_qubits_single_channel])

                # hardware-efficient CNOT on 8
                for i in range(0, num_wires_vqc, 2):
                    qml.CNOT(wires=[i, (i+1)%num_wires_vqc])
                for i in range(1, num_wires_vqc, 2):
                    qml.CNOT(wires=[i, (i+1)%num_wires_vqc])

                # qml.ctrl(qml.RX, (b % switch), control_values=(1))(theta_RY_switch[b], wires=switch)
                
                if all_layers_random_ctlrs is False:
                    if len(random_inds_list) != 1:
                        random_inds = sorted(np.random.choice(n_qubits - 1, size=kappa, replace=False))
                        random_inds_list.append(random_inds)
                        print("random_inds_list:", random_inds_list)

                    if type1_local_cost is True:
                        qml.ctrl(qml.Z, random_inds_list[0], control_values=control_vals)(wires=switch)
                    elif type2_local_cost is True:
                        for c_wire in random_inds_list[0]:
                            qml.ctrl(qml.Z, switch, control_values=(1))(wires=c_wire)
                else:
                    if len(random_inds_list) != num_vqc_blocks:
                        if weight_interaction_cost is True:
                            random_inds = sorted(np.random.choice(n_qubits_wo_blockwise - 1, size=kappa, replace=False))
                            for __ in range(n_qubits_wo_blockwise, n_qubits, 1):
                                random_inds.append(__)
                            # random_inds.append(n_qubits_wo_blockwise + b)
                        elif weight_interaction_cost__relaxed is True:
                            random_inds = sorted(np.random.choice(n_qubits_wo_blockwise - 1, size=kappa, replace=False))
                            random_inds.append(n_qubits_wo_blockwise + b)
                        else:
                            random_inds = sorted(np.random.choice(n_qubits - 1, size=kappa, replace=False))
                        random_inds_list.append(random_inds)
                        print("random_inds_list:", random_inds_list)

                    if weight_interaction_cost is True:
                        control_vals_temp = [0] * n_blocks
                        control_vals_temp[b] = 1
                        # print(control_vals, control_vals_temp)
                        control_vals.extend(control_vals_temp)
                        # control_vals.append(1)
                        
                        qml.ctrl(qml.U3, random_inds_list[b], control_values=control_vals)(*theta_RY_switch[b], wires=switch)
                        control_vals = control_vals[:-n_blocks]
                    elif weight_interaction_cost__relaxed is True:
                        control_vals.append(1)
                        
                        qml.ctrl(qml.U3, random_inds_list[b], control_values=control_vals)(*theta_RY_switch[b], wires=switch)
                        control_vals = control_vals[:-1]
                    else:
                        qml.ctrl(qml.U3, random_inds_list[b], control_values=control_vals)(*theta_RY_switch[b], wires=switch)

                    if type1_local_cost is True:
                        qml.ctrl(qml.Z, random_inds_list[b], control_values=control_vals)(wires=switch)
                    elif type2_local_cost is True:
                        for c_wire in random_inds_list[b]:
                            qml.ctrl(qml.Z, switch, control_values=(1))(wires=c_wire)
                
                # qml.ctrl(qml.U3, (b % switch), control_values=(1))(*theta_RY_switch[b], wires=switch)

        # final overlap
        if weight_interaction_cost is True:
            for _ in range(n_qubits_wo_blockwise, n_qubits, 1):
                qml.Hadamard(wires=_)
        elif weight_interaction_cost__relaxed is True:
            qml.ctrl(qml.Hadamard, [w for w in range(n_qubits_wo_blockwise, n_qubits, 1)], control_values=(1 for __ in range(n_qubits - n_qubits_wo_blockwise)))(wires=switch)
        else:
        # if True:
            qml.Hadamard(wires=switch)

        if weight_interaction_cost is True:
            # return [qml.expval(qml.PauliZ(wires=_)) for _ in range(n_qubits_wo_blockwise, n_qubits, 1)]
            return [qml.expval(qml.PauliZ(wires=_)) for _ in range(switch, n_qubits, 1)]
        else:
            return qml.expval(qml.PauliZ(wires=switch))

        # return qml.state(
    
    return qml.QNode(circuit, dev, interface="torch")


import os
from PIL import ImageOps, Image

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torchvision import transforms
from torch.utils.data import Dataset, DataLoader, IterableDataset

from sklearn.model_selection import train_test_split
from sklearn.utils import shuffle

import pandas as pd

import random
import csv
import itertools
import pickle

import matplotlib.pyplot as plt
import cv2

import numpy as np
import tqdm

import pennylane as qml

from sklearn import metrics
from sklearn.metrics import classification_report

def accuracy(distances, y, step=0.01):
    min_threshold_d = min(distances)
    max_threshold_d = max(distances)
    max_acc = 0
    best_thresh = 0
    same_id = (y == 1)

    for threshold_d in torch.arange(min_threshold_d, max_threshold_d+step, step):
        true_positive = (distances <= threshold_d) & (same_id)
        true_positive_rate = true_positive.sum().float() / same_id.sum().float()
        true_negative = (distances > threshold_d) & (~same_id)
        true_negative_rate = true_negative.sum().float() / (~same_id).sum().float()

        acc = 0.5 * (true_negative_rate + true_positive_rate)

        if acc > max_acc:
            max_acc = acc
            best_thresh = threshold_d
        # max_acc = max(max_acc, acc)
    return max_acc, best_thresh

def accuracy_triplet(anchor_embeddings, positive_embeddings, negative_embeddings, threshold=None):
    """
    Calculates the accuracy of a triplet network by comparing the distances.
    A triplet is considered "correct" if the distance from anchor to positive
    is less than the distance from anchor to negative.

    Args:
        anchor_embeddings (torch.Tensor): Embeddings of the anchor images.
        positive_embeddings (torch.Tensor): Embeddings of the positive images.
        negative_embeddings (torch.Tensor): Embeddings of the negative images.

    Returns:
        float: The accuracy as a percentage.
    """
    # Calculate the distances for all triplets
    dist_positive = torch.pairwise_distance(anchor_embeddings, positive_embeddings, p=2)
    dist_negative = torch.pairwise_distance(anchor_embeddings, negative_embeddings, p=2)

    # A prediction is correct if dist_positive < dist_negative
    correct_predictions = (dist_positive < dist_negative).sum().item()
    total_triplets = len(anchor_embeddings)

    return (correct_predictions / total_triplets) * 100

def accuracy_triplet__quantum(anchor_embeddings, dist_positive, dist_negative, threshold=None):
    """
    Calculates the accuracy of a triplet network by comparing the distances.
    A triplet is considered "correct" if the distance from anchor to positive
    is less than the distance from anchor to negative.

    Args:
        anchor_embeddings (torch.Tensor): Embeddings of the anchor images.
        positive_embeddings (torch.Tensor): Embeddings of the positive images.
        negative_embeddings (torch.Tensor): Embeddings of the negative images.

    Returns:
        float: The accuracy as a percentage.
    """
    # Calculate the distances for all triplets
    # dist_positive = torch.pairwise_distance(anchor_embeddings, positive_embeddings, p=2)
    # dist_negative = torch.pairwise_distance(anchor_embeddings, negative_embeddings, p=2)

    # A prediction is correct if dist_positive < dist_negative
    correct_predictions = (dist_positive > dist_negative).sum().item()
    total_triplets = len(anchor_embeddings)

    return (correct_predictions / total_triplets) * 100

def accuracy_contrastive_loss(distances, labels):
    """
    Calculates the optimal accuracy for a Siamese network trained with
    contrastive loss by finding the best threshold.

    Args:
        distances (torch.Tensor): A 1D tensor of distances between image pairs.
        labels (torch.Tensor): A 1D tensor of ground-truth labels (0 for similar, 1 for dissimilar).

    Returns:
        tuple: A tuple containing the maximum accuracy (float) and the
               corresponding best threshold (float).
    """
    max_accuracy = 0.0
    best_threshold = 0.0

    # Sort distances to iterate through all possible thresholds efficiently
    distances_sorted, _ = torch.sort(distances)

    for threshold in distances_sorted:
        # A pair is predicted as similar if its distance is <= threshold
        predictions = (distances <= threshold).float()

        # Calculate True Positives and True Negatives
        # True Positives: correct similar prediction (label=0, pred=0)
        # True Negatives: correct dissimilar prediction (label=1, pred=1)
        # predictions = 1 - predictions because label is 0 for similar, 1 for dissimilar
        tp = ((labels == 0) & (predictions == 1)).sum().item()
        tn = ((labels == 1) & (predictions == 0)).sum().item()

        accuracy = (tp + tn) / len(labels)

        if accuracy > max_accuracy:
            max_accuracy = accuracy
            best_threshold = threshold.item()

    return max_accuracy * 100, best_threshold